from sklearn.neighbors import NearestNeighbors
from sklearn.base import BaseEstimator
from copy import copy
from ._mdsw import _MDSW

import numpy as np
import scipy.sparse as sp


# inspired by implementation by Kuan-Hao Huang
# https://github.com/ej0cl6/csmlc


class CLEMS(BaseEstimator):
    """Embed the label space using a label network embedder from OpenNE

    Parameters
    ----------
    measure: Callable
        a cost function executed on two label vectors
    dimension: int
        the dimension of the label embedding vectors
    is_score: boolean
        set to True if measures is a score function (higher value is better), False if loss function (lower is better)
    param_dict: dict or None
        parameters passed to the embedder, don't use the dimension and graph parameters, this class will set them at fit


    Example code for using this embedder looks like this:

    .. code-block:: python

        from skmultilearn.embedding import CLEMS, EmbeddingClassifier
        from sklearn.ensemble import RandomForestRegressor
        from skmultilearn.adapt import MLkNN
        from sklearn.metrics import accuracy_score


        clf = EmbeddingClassifier(
            CLEMS(accuracy_score, True),
            RandomForestRegressor(n_estimators=10),
            MLkNN(k=5)
        )

        clf.fit(X_train, y_train)

        predictions = clf.predict(X_test)
    """

    def __init__(self, measure, is_score=False, params=None):
        self.measure = measure
        if is_score:
            self.measure = lambda x, y: 1 - measure(x, y)

        if params is None:
            params = {}

        self.params = params

    def fit(self, X, y):
        """Fits the embedder to data

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self
        """

        # get unique label combinations

        self.fit_transform(X, y)

    def fit_transform(self, X, y):
        """Fit the embedder and transform the output space

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        X, y_embedded
            results of the embedding, input and output space
        """

        if sp.issparse(y):
            idx = np.unique(y.tolil().rows, return_index=True)[1]
        else:
            idx = np.unique(y, axis=0, return_index=True)[1]

        y_unique = y[idx]
        n_unique = y_unique.shape[0]

        self.knn_ = NearestNeighbors(n_neighbors=1)
        self.knn_.fit(y_unique)

        nearest_points = self.knn_.kneighbors(y)[1][:, 0]
        nearest_points_counts = np.unique(nearest_points, return_counts=True)[1]

        # calculate delta matrix
        delta = np.zeros((2 * n_unique, 2 * n_unique))
        for i in range(n_unique):
            for j in range(n_unique):
                delta[i, n_unique + j] = np.sqrt(
                    self.measure(y_unique[None, i], y_unique[None, j])
                )
                delta[n_unique + j, i] = delta[i, n_unique + j]

        # calculate MDS embedding
        params = copy(self.params)
        params["n_components"] = y.shape[1]
        params["n_uq"] = n_unique
        params["uq_weight"] = nearest_points_counts
        params["dissimilarity"] = "precomputed"
        self.embedder_ = _MDSW(**params)

        y_unique_embedded = self.embedder_.fit(delta).embedding_
        y_unique_limited_to_before_trick = y_unique_embedded[n_unique:]

        knn_to_extend_embeddings_to_other_combinations = NearestNeighbors(n_neighbors=1)
        knn_to_extend_embeddings_to_other_combinations.fit(
            y_unique_limited_to_before_trick
        )
        neighboring_embeddings_indices = (
            knn_to_extend_embeddings_to_other_combinations.kneighbors(y)[1][:, 0]
        )

        return X, y_unique_embedded[neighboring_embeddings_indices]



# -*- coding: utf-8 -*-
"""Iterative stratification for multi-label data

The classifier follows methods outlined in Sechidis11 and Szymanski17 papers related to stratyfing
multi-label data.

In general what we expect from a given stratification output is that a strata, or a fold, is close to a given, demanded size,
usually equal to 1/k in k-fold approach, or a x% train to test set division in 2-fold splits.

The idea behind this stratification method is to assign label combinations to folds based on how much a given combination is desired
by a given fold, as more and more assignments are made, some folds are filled and positive evidence is directed into other folds,
in the end negative evidence is distributed based on a folds desirability of size.

You can also watch a `video presentation <http://videolectures.net/ecmlpkdd2011_tsoumakas_stratification/?q=stratification%20multi%20label>`_ by G. Tsoumakas which explains the algorithm. In 2017 Szymanski & Kajdanowicz extended the algorithm
to handle high-order relationships in the data set, if order = 1, the algorithm falls back to the original Sechidis11 setting.

If order is larger than 1 this class constructs a list of label combinations with replacement, i.e. allowing combinations of lower
order to be take into account. For example for combinations of order 2, the stratifier will consider both
label pairs (1, 2) and single labels denoted as (1,1) in the algorithm. In higher order cases the
when two combinations of different size have similar desirablity: the larger, i.e. more specific combination
is taken into consideration first, thus if a label pair (1,2) and label 1 represented as (1,1) are of similar
desirability, evidence for (1,2) will be assigned to folds first.

You can use this class exactly the same way you would use a normal scikit KFold class:

.. code-block :: python

    from skmultilearn.model_selection import IterativeStratification

    k_fold = IterativeStratification(n_splits=2, order=1):
    for train, test in k_fold.split(X, y):
        classifier.fit(X[train], y[train])
        result = classifier.predict(X[test])
        # do something with the result, comparing it to y[test]

Most of the methods of this class are private, you will not need them unless you are extending the method.

If you use this method to stratify data please cite both:
Sechidis, K., Tsoumakas, G., & Vlahavas, I. (2011). On the stratification of multi-label data. Machine Learning and Knowledge Discovery in Databases, 145-158.
http://lpis.csd.auth.gr/publications/sechidis-ecmlpkdd-2011.pdf

Piotr Szymański, Tomasz Kajdanowicz ; Proceedings of the First International Workshop on Learning with Imbalanced Domains: Theory and Applications, PMLR 74:22-35, 2017.
http://proceedings.mlr.press/v74/szyma%C5%84ski17a.html

Bibtex:

.. code-block:: bibtex

    @article{sechidis2011stratification,
      title={On the stratification of multi-label data},
      author={Sechidis, Konstantinos and Tsoumakas, Grigorios and Vlahavas, Ioannis},
      journal={Machine Learning and Knowledge Discovery in Databases},
      pages={145--158},
      year={2011},
      publisher={Springer}
    }

    @InProceedings{pmlr-v74-szymański17a,
      title =    {A Network Perspective on Stratification of Multi-Label Data},
      author =   {Piotr Szymański and Tomasz Kajdanowicz},
      booktitle =    {Proceedings of the First International Workshop on Learning with Imbalanced Domains: Theory and Applications},
      pages =    {22--35},
      year =     {2017},
      editor =   {Luís Torgo and Bartosz Krawczyk and Paula Branco and Nuno Moniz},
      volume =   {74},
      series =   {Proceedings of Machine Learning Research},
      address =      {ECML-PKDD, Skopje, Macedonia},
      publisher =    {PMLR},
    }
"""

from sklearn.model_selection._split import _BaseKFold
import numpy as np
import scipy.sparse as sp
import itertools
from sklearn.utils import check_random_state


def iterative_train_test_split(X, y, test_size, random_state=None):
    """Iteratively stratified train/test split

    Parameters
    ----------
    test_size : float, [0,1]
        the proportion of the dataset to include in the test split, the rest will be put in the train set

    random_state : None | int | np.random.RandomState
        the random state seed (optional)

    Returns
    -------
    X_train, y_train, X_test, y_test
        stratified division into train/test split
    """

    stratifier = IterativeStratification(
        n_splits=2,
        order=2,
        sample_distribution_per_fold=[test_size, 1.0 - test_size],
        random_state=random_state,
    )
    train_indexes, test_indexes = next(stratifier.split(X, y))

    X_train, y_train = X.loc[train_indexes], y.loc[train_indexes]
    X_test, y_test = X.loc[test_indexes], y.loc[test_indexes]

    return X_train, X_test, y_train, y_test


def _fold_tie_break(desired_samples_per_fold, M, random_state=check_random_state(None)):
    """Helper function to split a tie between folds with same desirability of a given sample

    Parameters
    ----------
    desired_samples_per_fold: np.array[Float], :code:`(n_splits)`
        number of samples desired per fold
    M : np.array(int)
        List of folds between which to break the tie

    random_state : None | int | np.random.RandomState
        the random state seed (optional)

    Returns
    -------
    fold_number : int
        The selected fold index to put samples into
    """
    if len(M) == 1:
        return M[0]
    else:
        max_val = max(desired_samples_per_fold[M])
        M_prim = np.where(np.array(desired_samples_per_fold) == max_val)[0]
        M_prim = np.array([x for x in M_prim if x in M])
        if random_state:
            if isinstance(random_state, np.random.RandomState):
                return random_state.choice(M_prim, 1)[0]
            else:
                np.random.seed(random_state)
        return np.random.choice(M_prim, 1)[0]


def _get_most_desired_combination(samples_with_combination):
    """Select the next most desired combination whose evidence should be split among folds

    Parameters
    ----------
    samples_with_combination : Dict[Combination, List[int]], :code:`(n_combinations)`
            map from each label combination present in y to list of sample indexes that have this combination assigned

    Returns
    -------
    combination: Combination
        the combination to split next
    """
    currently_chosen = None
    best_number_of_combinations, best_support_size = None, None

    for combination, evidence in samples_with_combination.items():
        number_of_combinations, support_size = (len(set(combination)), len(evidence))
        if support_size == 0:
            continue
        if currently_chosen is None or (
            best_number_of_combinations < number_of_combinations
            and best_support_size > support_size
        ):
            currently_chosen = combination
            best_number_of_combinations, best_support_size = (
                number_of_combinations,
                support_size,
            )

    return currently_chosen


class IterativeStratification(_BaseKFold):
    """Iteratively stratify a multi-label data set into folds

    Construct an interative stratifier that splits the data set into folds trying to maintain balanced representation
    with respect to order-th label combinations.

    Attributes
    ----------

    n_splits : number of splits, int
        the number of folds to stratify into

    order : int, >= 1
        the order of label relationship to take into account when balancing sample distribution across labels

    sample_distribution_per_fold : None or List[float], :code:`(n_splits)`
        desired percentage of samples in each of the folds, if None and equal distribution of samples per fold
        is assumed i.e. 1/n_splits for each fold. The value is held in :code:`self.percentage_per_fold`.

    shuffle : bool
        Whether to shuffle the data before splitting into batches. Note that the samples within each split
        will not be shuffled.

    random_state : int, RandomState instance or None
        integer to seed the RNG, or the RNG state to use; if None (the default), will use the global
        state of numpy RNG
    """

    def __init__(
        self,
        n_splits=3,
        order=1,
        sample_distribution_per_fold=None,
        shuffle=False,
        random_state=None,
    ):
        self._rng_state = check_random_state(random_state)
        need_shuffle = shuffle or random_state is not None
        self.order = order
        super(IterativeStratification, self).__init__(
            n_splits,
            shuffle=need_shuffle,
            random_state=self._rng_state if need_shuffle else None,
        )

        if sample_distribution_per_fold:
            self.percentage_per_fold = sample_distribution_per_fold
        else:
            self.percentage_per_fold = [
                1 / float(self.n_splits) for _ in range(self.n_splits)
            ]

    def _prepare_stratification(self, y):
        """Prepares variables for performing stratification

        For the purpose of clarity, the type Combination denotes List[int], :code:`(self.order)` and represents a
        label combination of the order we want to preserve among folds in stratification. The total number of
        combinations present in :code:`(y)` will be denoted as :code:`(n_combinations)`.

        Sets
        ----

        self.n_samples, self.n_labels : int, int
            shape of y

        self.desired_samples_per_fold: np.array[Float], :code:`(n_splits)`
            number of samples desired per fold

        self.desired_samples_per_combination_per_fold: Dict[Combination, np.array[Float]], :code:`(n_combinations, n_splits)`
            number of samples evidencing each combination desired per each fold

        Parameters
        ----------

        y : output matrix or array of arrays (n_samples, n_labels)

        Returns
        -------

        rows : List[List[int]], :code:`(n_samples, n_labels)`
            list of label indices assigned to each sample

        rows_used : Dict[int, bool], :code:`(n_samples)`
            boolean map from a given sample index to boolean value whether it has been already assigned to a fold or not

        all_combinations :  List[Combination], :code:`(n_combinations)`
            list of all label combinations of order self.order present in y

        per_row_combinations : List[Combination], :code:`(n_samples)`
            list of all label combinations of order self.order present in y per row

        samples_with_combination : Dict[Combination, List[int]], :code:`(n_combinations)`
            map from each label combination present in y to list of sample indexes that have this combination assigned

        folds: List[List[int]] (n_splits)
            list of lists to be populated with samples

        """
        self.n_samples, self.n_labels = y.shape
        self.desired_samples_per_fold = np.array(
            [self.percentage_per_fold[i] * self.n_samples for i in range(self.n_splits)]
        )
        rows = sp.lil_matrix(y).rows
        rows_used = {i: False for i in range(self.n_samples)}
        all_combinations = []
        per_row_combinations = [[] for i in range(self.n_samples)]
        samples_with_combination = {}
        folds = [[] for _ in range(self.n_splits)]

        # for every row
        for sample_index, label_assignment in enumerate(rows):
            # for every n-th order label combination
            # register combination in maps and lists used later
            for combination in itertools.combinations_with_replacement(
                label_assignment, self.order
            ):
                if combination not in samples_with_combination:
                    samples_with_combination[combination] = []

                samples_with_combination[combination].append(sample_index)
                all_combinations.append(combination)
                per_row_combinations[sample_index].append(combination)

        all_combinations = [list(x) for x in set(all_combinations)]

        self.desired_samples_per_combination_per_fold = {
            combination: np.array(
                [
                    len(evidence_for_combination) * self.percentage_per_fold[j]
                    for j in range(self.n_splits)
                ]
            )
            for combination, evidence_for_combination in samples_with_combination.items()
        }
        return (
            rows,
            rows_used,
            all_combinations,
            per_row_combinations,
            samples_with_combination,
            folds,
        )

    def _distribute_positive_evidence(
        self, rows_used, folds, samples_with_combination, per_row_combinations
    ):
        """Internal method to distribute evidence for labeled samples across folds

        For params, see documentation of :code:`self._prepare_stratification`. Does not return anything,
        modifies params.
        """
        l = _get_most_desired_combination(samples_with_combination)
        while l is not None:
            while len(samples_with_combination[l]) > 0:
                row = samples_with_combination[l].pop()
                if rows_used[row]:
                    continue

                max_val = max(self.desired_samples_per_combination_per_fold[l])
                M = np.where(
                    np.array(self.desired_samples_per_combination_per_fold[l])
                    == max_val
                )[0]
                m = _fold_tie_break(
                    self.desired_samples_per_combination_per_fold[l], M, self._rng_state
                )
                folds[m].append(row)
                rows_used[row] = True
                for i in per_row_combinations[row]:
                    if row in samples_with_combination[i]:
                        samples_with_combination[i].remove(row)
                    self.desired_samples_per_combination_per_fold[i][m] -= 1
                self.desired_samples_per_fold[m] -= 1

            l = _get_most_desired_combination(samples_with_combination)

    def _distribute_negative_evidence(self, rows_used, folds):
        """Internal method to distribute evidence for unlabeled samples across folds

        For params, see documentation of :code:`self._prepare_stratification`. Does not return anything,
        modifies params.
        """
        available_samples = [i for i, v in rows_used.items() if not v]
        samples_left = len(available_samples)

        while samples_left > 0:
            row = available_samples.pop()
            rows_used[row] = True
            samples_left -= 1
            fold_selected = self._rng_state.choice(
                np.where(self.desired_samples_per_fold > 0)[0], 1
            )[0]
            self.desired_samples_per_fold[fold_selected] -= 1
            folds[fold_selected].append(row)

    def _iter_test_indices(self, X, y=None, groups=None):
        """Internal method for providing scikit-learn's split with folds

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
            Note that providing ``y`` is sufficient to generate the splits and
            hence ``np.zeros(n_samples)`` may be used as a placeholder for
            ``X`` instead of actual training data.
        y : array-like, shape (n_samples,)
            The target variable for supervised learning problems.
            Stratification is done based on the y labels.
        groups : object
            Always ignored, exists for compatibility.

        Yields
        ------
        fold : List[int]
            indexes of test samples for a given fold, yielded for each of the folds
        """
        (
            rows,
            rows_used,
            all_combinations,
            per_row_combinations,
            samples_with_combination,
            folds,
        ) = self._prepare_stratification(y)

        self._distribute_positive_evidence(
            rows_used, folds, samples_with_combination, per_row_combinations
        )
        self._distribute_negative_evidence(rows_used, folds)

        for fold in folds:
            yield fold



from builtins import range
from ..base import MLClassifierBase
from ..utils import get_matrix_in_format

from sklearn.neighbors import NearestNeighbors
import numpy as np
import scipy.sparse as sparse


class MLkNN(MLClassifierBase):
    """kNN classification method adapted for multi-label classification

    MLkNN builds uses k-NearestNeighbors find nearest examples to a test class and uses Bayesian inference
    to select assigned labels.

    Parameters
    ----------
    k : int
        number of neighbours of each input instance to take into account
    s: float (default is 1.0)
        the smoothing parameter
    ignore_first_neighbours : int (default is 0)
        ability to ignore first N neighbours, useful for comparing
        with other classification software.
    n_jobs: int or None, optional (default=None)
        The number of parallel jobs to run for neighbors search. None means 1 unless in a joblib.parallel_backend context. -1 means using all processors.

    Attributes
    ----------
    knn_ : an instance of sklearn.NearestNeighbors
        the nearest neighbors single-label classifier used underneath



    .. note:: If you don't know what :code:`ignore_first_neighbours`
              does, the default is safe. Please see this `issue`_.


    .. _issue: https://github.com/scikit-multilearn/scikit-multilearn/issues/22


    References
    ----------

    If you use this classifier please cite the original paper introducing the method:

    .. code :: bibtex

        @article{zhang2007ml,
          title={ML-KNN: A lazy learning approach to multi-label learning},
          author={Zhang, Min-Ling and Zhou, Zhi-Hua},
          journal={Pattern recognition},
          volume={40},
          number={7},
          pages={2038--2048},
          year={2007},
          publisher={Elsevier}
        }

    Examples
    --------

    Here's a very simple example of using MLkNN with a fixed number of neighbors:

    .. code :: python

        from skmultilearn.adapt import MLkNN

        classifier = MLkNN(k=3)

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)


    You can also use :class:`~sklearn.model_selection.GridSearchCV` to find an optimal set of parameters:

    .. code :: python

        from skmultilearn.adapt import MLkNN
        from sklearn.model_selection import GridSearchCV

        parameters = {'k': range(1,3), 's': [0.5, 0.7, 1.0]}
        score = 'f1_macro'

        clf = GridSearchCV(MLkNN(), parameters, scoring=score)
        clf.fit(X, y)

        print (clf.best_params_, clf.best_score_)

        # output
        ({'k': 1, 's': 0.5}, 0.78988303374297597)

    """

    def __init__(self, k=10, s=1.0, ignore_first_neighbours=0, n_jobs=None):
        """Initializes the classifier

        Parameters
        ----------
        k : int
            number of neighbours of each input instance to take into account
        s: float (default is 1.0)
            the smoothing parameter
        ignore_first_neighbours : int (default is 0)
            ability to ignore first N neighbours, useful for comparing
            with other classification software.
        n_jobs: int or None, optional (default=None)
            The number of parallel jobs to run for neighbors search. None means 1 unless in a joblib.parallel_backend context. -1 means using all processors.


        Attributes
        ----------
        knn_ : an instance of sklearn.NearestNeighbors
            the nearest neighbors single-label classifier used underneath

        .. note:: If you don't know what :code:`ignore_first_neighbours`
                  does, the default is safe. Please see this `issue`_.

        .. _issue: https://github.com/scikit-multilearn/scikit-multilearn/issues/22
        """
        super(MLkNN, self).__init__()
        self.k = k  # Number of neighbours
        self.s = s  # Smooth parameter
        self.ignore_first_neighbours = ignore_first_neighbours
        self.n_jobs = n_jobs
        self.knn_ = NearestNeighbors(n_neighbors=self.k, n_jobs=self.n_jobs)
        self.copyable_attrs = ["k", "s", "ignore_first_neighbours", "n_jobs"]

    def _compute_prior(self, y):
        """Helper function to compute for the prior probabilities

        Parameters
        ----------
        y : numpy.ndarray or scipy.sparse
            the training labels

        Returns
        -------
        numpy.ndarray
            the prior probability given true
        numpy.ndarray
            the prior probability given false
        """
        prior_prob_true = np.array(
            (self.s + y.sum(axis=0)) / (self.s * 2 + self._num_instances)
        )[0]
        prior_prob_false = 1 - prior_prob_true

        return (prior_prob_true, prior_prob_false)

    def _compute_cond(self, X, y):
        """Helper function to compute for the posterior probabilities

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse
            input features, can be a dense or sparse matrix of size
            :code:`(n_samples, n_features)`
        y : numpy.ndaarray or scipy.sparse {0,1}
            binary indicator matrix with label assignments.

        Returns
        -------
        numpy.ndarray
            the posterior probability given true
        numpy.ndarray
            the posterior probability given false
        """

        self.knn_.fit(X)
        c = sparse.lil_matrix((self._num_labels, self.k + 1), dtype="i8")
        cn = sparse.lil_matrix((self._num_labels, self.k + 1), dtype="i8")

        label_info = get_matrix_in_format(y, "dok")

        neighbors = [
            a[self.ignore_first_neighbours :]
            for a in self.knn_.kneighbors(
                X, self.k + self.ignore_first_neighbours, return_distance=False
            )
        ]

        for instance in range(self._num_instances):
            deltas = label_info[neighbors[instance], :].sum(axis=0)
            for label in range(self._num_labels):
                if label_info[instance, label] == 1:
                    c[label, deltas[0, label]] += 1
                else:
                    cn[label, deltas[0, label]] += 1

        c_sum = c.sum(axis=1)
        cn_sum = cn.sum(axis=1)

        cond_prob_true = sparse.lil_matrix(
            (self._num_labels, self.k + 1), dtype="float"
        )
        cond_prob_false = sparse.lil_matrix(
            (self._num_labels, self.k + 1), dtype="float"
        )
        for label in range(self._num_labels):
            for neighbor in range(self.k + 1):
                cond_prob_true[label, neighbor] = (self.s + c[label, neighbor]) / (
                    self.s * (self.k + 1) + c_sum[label, 0]
                )
                cond_prob_false[label, neighbor] = (self.s + cn[label, neighbor]) / (
                    self.s * (self.k + 1) + cn_sum[label, 0]
                )
        return cond_prob_true, cond_prob_false

    def fit(self, X, y):
        """Fit classifier with training data

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse
            input features, can be a dense or sparse matrix of size
            :code:`(n_samples, n_features)`
        y : numpy.ndaarray or scipy.sparse {0,1}
            binary indicator matrix with label assignments.

        Returns
        -------
        self
            fitted instance of self
        """

        self._label_cache = get_matrix_in_format(y, "lil")
        self._num_instances = self._label_cache.shape[0]
        self._num_labels = self._label_cache.shape[1]
        # Computing the prior probabilities
        self._prior_prob_true, self._prior_prob_false = self._compute_prior(
            self._label_cache
        )
        # Computing the posterior probabilities
        self._cond_prob_true, self._cond_prob_false = self._compute_cond(
            X, self._label_cache
        )
        return self

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse matrix of int
            binary indicator matrix with label assignments with shape
            :code:`(n_samples, n_labels)`
        """

        result = sparse.lil_matrix((X.shape[0], self._num_labels), dtype="i8")
        neighbors = [
            a[self.ignore_first_neighbours :]
            for a in self.knn_.kneighbors(
                X, self.k + self.ignore_first_neighbours, return_distance=False
            )
        ]
        for instance in range(X.shape[0]):
            deltas = self._label_cache[neighbors[instance],].sum(axis=0)

            for label in range(self._num_labels):
                p_true = (
                    self._prior_prob_true[label]
                    * self._cond_prob_true[label, deltas[0, label]]
                )
                p_false = (
                    self._prior_prob_false[label]
                    * self._cond_prob_false[label, deltas[0, label]]
                )
                result[instance, label] = int(p_true >= p_false)

        return result

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse matrix of int
            binary indicator matrix with label assignment probabilities
            with shape :code:`(n_samples, n_labels)`
        """
        result = sparse.lil_matrix((X.shape[0], self._num_labels), dtype="float")
        neighbors = [
            a[self.ignore_first_neighbours :]
            for a in self.knn_.kneighbors(
                X, self.k + self.ignore_first_neighbours, return_distance=False
            )
        ]
        for instance in range(X.shape[0]):
            deltas = self._label_cache[neighbors[instance],].sum(axis=0)

            for label in range(self._num_labels):
                p_true = (
                    self._prior_prob_true[label]
                    * self._cond_prob_true[label, deltas[0, label]]
                )
                p_false = (
                    self._prior_prob_false[label]
                    * self._cond_prob_false[label, deltas[0, label]]
                )
                result[instance, label] = p_true / (p_true + p_false)

        return result



# Authors: Grzegorz Kulakowski <grzegorz7w@gmail.com>
# License: BSD 3 clause
from skmultilearn.base import MLClassifierBase

import numpy as np
import scipy.sparse as sp
from scipy.linalg import norm
from scipy.sparse.linalg import inv as inv_sparse
from scipy.linalg import inv as inv_dense


class MLTSVM(MLClassifierBase):
    """Twin multi-Label Support Vector Machines

    Parameters
    ----------
    c_k : int
        the empirical risk penalty parameter that determines the trade-off between the loss terms
    sor_omega: float (default is 1.0)
        the smoothing parameter
    threshold : int (default is 1e-6)
            threshold above which a label should be assigned
    lambda_param : float (default is 1.0)
            the regularization parameter
    max_iteration : int (default is 500)
            maximum number of iterations to use in successive overrelaxation


    References
    ----------

    If you use this classifier please cite the original paper introducing the method:

    .. code :: bibtex

        @article{chen2016mltsvm,
          title={MLTSVM: a novel twin support vector machine to multi-label learning},
          author={Chen, Wei-Jie and Shao, Yuan-Hai and Li, Chun-Na and Deng, Nai-Yang},
          journal={Pattern Recognition},
          volume={52},
          pages={61--74},
          year={2016},
          publisher={Elsevier}
        }


    Examples
    --------

    Here's a very simple example of using MLTSVM with a fixed number of neighbors:

    .. code :: python

        from skmultilearn.adapt import MLTSVM

        classifier = MLTSVM(c_k = 2**-1)

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)


    You can also use :class:`~sklearn.model_selection.GridSearchCV` to find an optimal set of parameters:

    .. code :: python

        from skmultilearn.adapt import MLTSVM
        from sklearn.model_selection import GridSearchCV

        parameters = {'c_k': [2**i for i in range(-5, 5, 2)]}
        score = 'f1-macro

        clf = GridSearchCV(MLTSVM(), parameters, scoring=score)
        clf.fit(X, y)

        print (clf.best_params_, clf.best_score_)

        # output
        {'c_k': 0.03125} 0.347518217573


    """

    def __init__(
        self, c_k=0, sor_omega=1.0, threshold=1e-6, lambda_param=1.0, max_iteration=500
    ):
        super(MLClassifierBase, self).__init__()
        self.max_iteration = max_iteration
        self.threshold = threshold
        self.lambda_param = lambda_param  # TODO: possibility to add different lambda to different labels
        self.c_k = c_k
        self.sor_omega = sor_omega
        self.copyable_attrs = [
            "c_k",
            "sor_omega",
            "lambda_param",
            "threshold",
            "max_iteration",
        ]

    def fit(self, X, Y):
        n_labels = Y.shape[1]
        m = X.shape[1]  # Count of features
        self.wk_bk = np.zeros([n_labels, m + 1], dtype=float)

        if sp.issparse(X):
            identity_matrix = sp.identity(m + 1)
            _inv = inv_sparse
        else:
            identity_matrix = np.identity(m + 1)
            _inv = inv_dense

        X_bias = _hstack(X, np.ones((X.shape[0], 1), dtype=X.dtype))
        self.iteration_count = []
        for label in range(0, n_labels):
            # Calculate the parameter Q for overrelaxation
            H_k = _get_x_class_instances(X_bias, Y, label)
            G_k = _get_x_noclass_instances(X_bias, Y, label)
            Q_knoPrefixGk = _inv(
                (H_k.T).dot(H_k) + self.lambda_param * identity_matrix
            ).dot(G_k.T)
            Q_k = G_k.dot(Q_knoPrefixGk)
            Q_k = (Q_k + Q_k.T) / 2.0

            # Calculate other
            alpha_k = self._successive_overrelaxation(self.sor_omega, Q_k)
            if sp.issparse(X):
                self.wk_bk[label] = -Q_knoPrefixGk.dot(alpha_k).T
            else:
                self.wk_bk[label] = (-np.dot(Q_knoPrefixGk, alpha_k)).T

        self.wk_norms = norm(self.wk_bk, axis=1)
        self.treshold = 1.0 / np.max(self.wk_norms)

    def predict(self, X):
        X_with_bias = _hstack(X, np.ones((X.shape[0], 1), dtype=X.dtype))
        wk_norms_multiplicated = self.wk_norms[
            np.newaxis, :
        ]  # change to form [[wk1, wk2, ..., wkk]]
        all_distances = (-X_with_bias.dot(self.wk_bk.T)) / wk_norms_multiplicated
        predicted_y = np.where(all_distances < self.treshold, 1, 0)
        # TODO: It's possible to add condition to: add label if no labels is in row.
        return predicted_y

    def _successive_overrelaxation(self, omegaW, Q):
        # Initialization
        D = np.diag(Q)  # Only one dimension vector - is enough
        D_inv = 1.0 / D  # D-1 simplify form
        small_l = Q.shape[1]
        oldnew_alpha = np.zeros([small_l, 1])  # buffer

        is_not_enough = True
        was_going_down = False
        last_alfa_norm_change = -1

        nr_iter = 0
        while is_not_enough:  # do while
            oldAlpha = oldnew_alpha
            for j in range(0, small_l):  # It's from last alpha to first
                oldnew_alpha[j] = oldAlpha[j] - omegaW * D_inv[j] * (
                    Q[j, :].T.dot(oldnew_alpha) - 1
                )
            oldnew_alpha = oldnew_alpha.clip(0.0, self.c_k)
            alfa_norm_change = norm(oldnew_alpha - oldAlpha)

            if not was_going_down and last_alfa_norm_change > alfa_norm_change:
                was_going_down = True
            is_not_enough = (
                alfa_norm_change > self.threshold
                and nr_iter < self.max_iteration
                and ((not was_going_down) or last_alfa_norm_change > alfa_norm_change)
            )
            # TODO: maybe add any(oldnew_alpha != oldAlpha)

            last_alfa_norm_change = alfa_norm_change
            nr_iter += 1
        self.iteration_count.append(nr_iter)
        return oldnew_alpha


def _get_x_noclass_instances(X, Y, label_class):
    if sp.issparse(Y):
        indices = np.where(Y[:, 1].A == 0)[0]
    else:
        indices = np.where(Y[:, 1] == 0)[0]
    return X[indices, :]


def _get_x_class_instances(X, Y, label_class):
    if sp.issparse(Y):
        indices = Y[:, label_class].nonzero()[0]
    else:
        indices = np.nonzero(Y[:, label_class])[0]
    return X[indices, :]


def _hstack(X, Y):
    if sp.issparse(X):
        return sp.hstack([X, Y], format=X.format)
    else:
        return np.hstack([X, Y])



from __future__ import absolute_import
from .base import LabelSpaceClustererBase


class FixedLabelSpaceClusterer(LabelSpaceClustererBase):
    """Return a fixed label space partition

    This clusterer takes a predefined fixed ``clustering`` of the label space and returns it in fit_predict as the label
    space division. This is useful for employing expert knowledge about label space division or partitions in ensemble
    classifiers such as: :class:`~skmultilearn.ensemble.LabelSpacePartitioningClassifier` or
    :class:`~skmultilearn.ensemble.MajorityVotingClassifier`.

    Parameters
    ----------
    clusters : array of arrays of int
        provided partition of the label space in the for of numpy array of
        numpy arrays of indexes for each partition, ex. ``[[0,1],[2,3]]``


    An example use of the fixed clusterer with a label partitioning classifier to train randomforests for a set of
    subproblems defined upon expert knowledge:

    .. code :: python

        from skmultilearn.ensemble import LabelSpacePartitioningClassifier
        from skmultilearn.cluster import FixedLabelSpaceClusterer
        from skmultilearn.problem_transform import LabelPowerset
        from sklearn.ensemble import RandomForestClassifier

        classifier = LabelSpacePartitioningClassifier(
            classifier = LabelPowerset(
                classifier=RandomForestClassifier(n_estimators=100),
                require_dense = [False, True]
            ),
            require_dense = [True, True],
            clusterer = FixedLabelSpaceClusterer(clustering=[[1,2,3], [0,4]])
        )

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)

    """

    def __init__(self, clusters=None):
        super(FixedLabelSpaceClusterer, self).__init__()

        self.clusters = clusters

    def fit_predict(self, X, y):
        """Returns the provided label space division

        Parameters
        ----------
        X : None
            currently unused, left for scikit compatibility
        y : scipy.sparse
            label space of shape :code:`(n_samples, n_labels)`

        Returns
        -------
        arrray of arrays of label indexes (numpy.ndarray)
            label space division, each sublist represents labels that are in that community
        """

        return self.clusters



import copy
import numpy as np

from scipy.sparse import hstack, issparse, lil_matrix

from ..base.problem_transformation import ProblemTransformationBase
from ..base.base import MLClassifierBase


class BinaryRelevance(ProblemTransformationBase):
    """Performs classification per label

    Transforms a multi-label classification problem with L labels
    into L single-label separate binary classification problems
    using the same base classifier provided in the constructor. The
    prediction output is the union of all per label classifiers

    Parameters
    ----------
    classifier : :class:`~sklearn.base.BaseEstimator`
        scikit-learn compatible base classifier
    require_dense : [bool, bool], optional
        whether the base classifier requires dense representations
        for input features and classes/labels matrices in fit/predict.
        If value not provided, sparse representations are used if base classifier is
        an instance of :class:`~skmultilearn.base.MLClassifierBase` and dense otherwise.

    Attributes
    ----------
    model_count_ : int
        number of trained models, in this classifier equal to `n_labels`
    partition_ : List[List[int]], shape=(`model_count_`,)
        list of lists of label indexes, used to index the output space matrix, set in :meth:`_generate_partition`
        via :meth:`fit`
    classifiers_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `model_count`
        list of classifiers trained per partition, set in :meth:`fit`

    Notes
    -----
    .. note ::

        This is one of the most basic approaches to multi-label classification, it ignores relationships between labels.

    Examples
    --------
    An example use case for Binary Relevance classification
    with an :class:`sklearn.svm.SVC` base classifier which supports sparse input:


    .. code-block:: python

        from skmultilearn.problem_transform import BinaryRelevance
        from sklearn.svm import SVC

        # initialize Binary Relevance multi-label classifier
        # with an SVM classifier
        # SVM in scikit only supports the X matrix in sparse representation

        classifier = BinaryRelevance(
            classifier = SVC(),
            require_dense = [False, True]
        )

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)

    Another way to use this classifier is to select the best scenario from a set of single-label classifiers used
    with Binary Relevance, this can be done using cross validation grid search. In the example below, the model
    with highest accuracy results is selected from either a :class:`sklearn.naive_bayes.MultinomialNB` or
    :class:`sklearn.svm.SVC` base classifier, alongside with best parameters for that base classifier.

    .. code-block:: python

        from skmultilearn.problem_transform import BinaryRelevance
        from sklearn.model_selection import GridSearchCV
        from sklearn.naive_bayes import MultinomialNB
        from sklearn.svm import SVC

        parameters = [
            {
                'classifier': [MultinomialNB()],
                'classifier__alpha': [0.7, 1.0],
            },
            {
                'classifier': [SVC()],
                'classifier__kernel': ['rbf', 'linear'],
            },
        ]


        clf = GridSearchCV(BinaryRelevance(), parameters, scoring='accuracy')
        clf.fit(x, y)

        print (clf.best_params_, clf.best_score_)

        # result:
        #
        # {
        #   'classifier': SVC(C=1.0, cache_size=200, class_weight=None, coef0=0.0,
        #   decision_function_shape='ovr', degree=3, gamma='auto', kernel='linear',
        #   max_iter=-1, probability=False, random_state=None, shrinking=True,
        #   tol=0.001, verbose=False), 'classifier__kernel': 'linear'
        # } 0.17

    """

    def __init__(self, classifier=None, require_dense=None):
        super(BinaryRelevance, self).__init__(classifier, require_dense)

    def _generate_partition(self, X, y):
        """Partitions the label space into singletons

        Sets `self.partition_` (list of single item lists) and `self.model_count_` (equal to number of labels).

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            not used, only for API compatibility
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `int`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        self.partition_ = list(range(y.shape[1]))
        self.model_count_ = y.shape[1]

    def fit(self, X, y):
        """Fits classifier to training data

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self

        Notes
        -----
        .. note :: Input matrices are converted to sparse format internally if a numpy representation is passed
        """
        X = self._ensure_input_format(X, sparse_format="csr", enforce_sparse=True)
        y = self._ensure_output_format(y, sparse_format="csc", enforce_sparse=True)

        self.classifiers_ = []
        self._generate_partition(X, y)
        self._label_count = y.shape[1]

        for i in range(self.model_count_):
            classifier = copy.deepcopy(self.classifier)
            y_subset = self._generate_data_subset(y, self.partition_[i], axis=1)
            if issparse(y_subset) and y_subset.ndim > 1 and y_subset.shape[1] == 1:
                y_subset = np.ravel(y_subset.toarray())
            classifier.fit(
                self._ensure_input_format(X), self._ensure_output_format(y_subset)
            )
            self.classifiers_.append(classifier)

        return self

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        predictions = [
            self._ensure_multi_label_from_single_class(
                self.classifiers_[label].predict(self._ensure_input_format(X))
            )
            for label in range(self.model_count_)
        ]

        return hstack(predictions)

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `float in [0.0, 1.0]`, shape=(n_samples, n_labels)
            matrix with label assignment probabilities
        """

        result = lil_matrix((X.shape[0], self._label_count), dtype="float")
        for label_assignment, classifier in zip(self.partition_, self.classifiers_):
            if isinstance(self.classifier, MLClassifierBase):
                # the multilabel classifier should provide a (n_samples, n_labels) matrix
                # we just need to reorder it column wise
                result[:, label_assignment] = classifier.predict_proba(X)
            else:
                # a base classifier for binary relevance returns
                # n_samples x n_classes, where n_classes = [0, 1] - 1 is the probability of
                # the label being assigned
                result[
                    :, label_assignment
                ] = self._ensure_multi_label_from_single_class(
                    classifier.predict_proba(self._ensure_input_format(X))
                )[
                    :, 1
                ]  # probability that label is assigned

        return result



"""
The :mod:`skmultilearn.base` module implements base
classifier classes for scikit-multilearn's multi-label classification.

Two base classifier classes are in use currently in scikit-multilearn:

- :class:`MLClassifierBase` - a generic base class for multi-label classifiers
- :class:`ProblemTransformationBase` - the base class for problem transformation and ensemble approaches that handles a base classifier
"""

from .base import MLClassifierBase
from .problem_transformation import ProblemTransformationBase

__all__ = ["MLClassifierBase", "ProblemTransformationBase"]



import numpy as np
import scipy.sparse as sp

SPARSE_FORMAT_TO_CONSTRUCTOR = {
    "bsr": sp.bsr_matrix,
    "coo": sp.coo_matrix,
    "csc": sp.csc_matrix,
    "csr": sp.csr_matrix,
    "dia": sp.dia_matrix,
    "dok": sp.dok_matrix,
    "lil": sp.lil_matrix,
}


def get_matrix_in_format(original_matrix, matrix_format):
    """Converts matrix to format

    Parameters
    ----------

    original_matrix : np.matrix or scipy matrix or np.array of np. arrays
        matrix to convert

    matrix_format : string
        format

    Returns
    -------

    matrix : scipy matrix
        matrix in given format
    """
    if isinstance(original_matrix, np.ndarray):
        return SPARSE_FORMAT_TO_CONSTRUCTOR[matrix_format](original_matrix)

    if original_matrix.getformat() == matrix_format:
        return original_matrix

    return original_matrix.asformat(matrix_format)


def matrix_creation_function_for_format(sparse_format):
    if sparse_format not in SPARSE_FORMAT_TO_CONSTRUCTOR:
        return None

    return SPARSE_FORMAT_TO_CONSTRUCTOR[sparse_format]


def measure_per_label(measure, y_true, y_predicted):
    """
    Return per label results of a scikit-learn compatible quality measure

    Parameters
    ----------
    measure : callable
        scikit-compatible quality measure function
    y_true : sparse matrix
        ground truth
    y_predicted : sparse matrix
        the predicted result

    Returns
    -------
    List[int or float]
        score from a given measure depending on what the measure returns
    """
    return [
        measure(y_true[:, i].toarray(), y_predicted[:, i].toarray())
        for i in range(y_true.shape[1])
    ]



from keras.wrappers.scikit_learn import KerasClassifier
from sklearn.base import BaseEstimator
from copy import copy


class Keras(BaseEstimator):
    def __init__(self, build_function, multi_class=False, keras_params=None):
        if not callable(build_function):
            raise ValueError("Model construction function must be callable.")

        self.multi_class = multi_class
        self.build_function = build_function
        if keras_params is None:
            keras_params = {}

        self.keras_params = keras_params

    def fit(self, X, y):
        if self.multi_class:
            self.n_classes_ = len(set(y))
        else:
            self.n_classes_ = 1

        build_callable = lambda: self.build_function(X.shape[1], self.n_classes_)
        keras_params = copy(self.keras_params)
        keras_params["build_fn"] = build_callable

        self.classifier_ = KerasClassifier(**keras_params)
        self.classifier_.fit(X, y)

    def predict(self, X):
        return self.classifier_.predict(X)



# -*- coding: utf-8 -*-
from ..problem_transform.br import BinaryRelevance

from scipy import sparse


class LabelSpacePartitioningClassifier(BinaryRelevance):
    """Partition label space and classify each subspace separately


    This classifier performs classification by:

    1. partitioning the label space into separate, smaller multi-label sub problems, using the supplied label
    space clusterer

    2. training an instance of the supplied base mult-label classifier for each label space subset in the partition

    3. predicting the result with each of subclassifiers and returning the sum of their results

    Parameters
    ----------
    classifier : :class:`~sklearn.base.BaseEstimator`
        the base classifier that will be used in a class, will be
        automatically put under :code:`self.classifier`.
    clusterer : :class:`~skmultilearn.cluster.LabelSpaceClustererBase`
        object that partitions the output space, will be
        automatically put under :code:`self.clusterer`.
    require_dense : [bool, bool]
        whether the base classifier requires [input, output] matrices
        in dense representation, will be automatically
        put under :code:`self.require_dense`.

    Attributes
    ----------
    model_count_ : int
        number of trained models, in this classifier equal to the number of partitions
    partition_ : List[List[int]], shape=(`model_count_`,)
        list of lists of label indexes, used to index the output space matrix, set in :meth:`_generate_partition`
        via :meth:`fit`
    classifiers : List[:class:`~sklearn.base.BaseEstimator`], shape=(`model_count_`,)
        list of classifiers trained per partition, set in :meth:`fit`


    References
    ----------

    If you use this clusterer please cite the clustering paper:

    .. code:: latex

        @Article{datadriven,
            author = {Szymański, Piotr and Kajdanowicz, Tomasz and Kersting, Kristian},
            title = {How Is a Data-Driven Approach Better than Random Choice in
            Label Space Division for Multi-Label Classification?},
            journal = {Entropy},
            volume = {18},
            year = {2016},
            number = {8},
            article_number = {282},
            url = {http://www.mdpi.com/1099-4300/18/8/282},
            issn = {1099-4300},
            doi = {10.3390/e18080282}
        }

    Examples
    --------
    Here's an example of building a partitioned ensemble of Classifier Chains

    .. code :: python

        from skmultilearn.ensemble import MajorityVotingClassifier
        from skmultilearn.cluster import FixedLabelSpaceClusterer
        from skmultilearn.problem_transform import ClassifierChain
        from sklearn.naive_bayes import GaussianNB


        classifier = MajorityVotingClassifier(
            clusterer = FixedLabelSpaceClusterer(clusters = [[1,3,4], [0, 2, 5]]),
            classifier = ClassifierChain(classifier=GaussianNB())
        )
        classifier.fit(X_train,y_train)
        predictions = classifier.predict(X_test)

    More advanced examples can be found in `the label relations exploration guide <../labelrelations.ipynb>`_

    """

    def __init__(self, classifier=None, clusterer=None, require_dense=None):
        super(LabelSpacePartitioningClassifier, self).__init__(
            classifier, require_dense
        )
        self.clusterer = clusterer
        self.copyable_attrs = ["clusterer", "classifier", "require_dense"]

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse of int
            binary indicator matrix with label assignments with shape
            :code:`(n_samples, n_labels)`
        """
        X = self._ensure_input_format(X, sparse_format="csr", enforce_sparse=True)
        result = sparse.lil_matrix((X.shape[0], self._label_count), dtype=int)

        for model in range(self.model_count_):
            predictions = self._ensure_output_format(
                self.classifiers_[model].predict(X),
                sparse_format=None,
                enforce_sparse=True,
            ).nonzero()
            for row, column in zip(predictions[0], predictions[1]):
                result[row, self.partition_[model][column]] = 1

        return result

    def _generate_partition(self, X, y):
        """Cluster the label space

        Saves the partiton generated by the clusterer to :code:`self.partition_` and
        sets :code:`self.model_count_` to number of clusers and :code:`self._label_count`
        to number of labels.

        Parameters
        -----------
        X : numpy.ndarray or scipy.sparse
            input features of shape :code:`(n_samples, n_features)`, passed to clusterer
        y : numpy.ndarray or scipy.sparse
            binary indicator matrix with label assigments of shape
            :code:`(n_samples, n_labels)`

        Returns
        -------
        LabelSpacePartitioningClassifier
            returns an instance of itself
        """

        self.partition_ = self.clusterer.fit_predict(X, y)
        self.model_count_ = len(self.partition_)
        self._label_count = y.shape[1]

        return self



from copy import copy
from openne.gf import GraphFactorization
from openne.graph import Graph
from openne.grarep import GraRep
from openne.hope import HOPE
from openne.lap import LaplacianEigenmaps
from openne.line import LINE
from openne.lle import LLE
import networkx as nx
import numpy as np
import tensorflow as tf
import scipy.sparse as sp


class OpenNetworkEmbedder:
    """Embed the label space using a label network embedder from OpenNE

    Implements an OpenNE based LNEMLC: label network embeddings for multi-label classification.

    Parameters
    ----------
    graph_builder: a GraphBuilderBase inherited transformer
        the graph builder to provide the adjacency matrix and weight map for the underlying graph
    embedding : string, one of {'GraphFactorization', 'GraRep', 'HOPE', 'LaplacianEigenmaps', 'LINE', 'LLE'}
        the selected OpenNE_ embedding

        +----------------------+--------------------------------------------------------------------------------+
        | Method name string   |                             Description                                        |
        +----------------------+--------------------------------------------------------------------------------+
        | GraphFactorization_  | Graph factorization embeddings                                                 |
        +----------------------+--------------------------------------------------------------------------------+
        | GraRep_              | Graph representations with global structural information                       |
        +----------------------+--------------------------------------------------------------------------------+
        | HOPE_                | High-order Proximity Preserved Embedding                                       |
        +----------------------+--------------------------------------------------------------------------------+
        | LaplacianEigenmaps_  | Detecting communities from multiple async label propagation on the graph       |
        +----------------------+--------------------------------------------------------------------------------+
        | LINE_                | Large-scale information network embedding                                      |
        +----------------------+--------------------------------------------------------------------------------+
        | LLE_                 | Locally Linear Embedding                                                       |
        +----------------------+--------------------------------------------------------------------------------+

        .. _OpenNE: https://github.com/thunlp/OpenNE/
        .. _GraphFactorization: https://github.com/thunlp/OpenNE/blob/master/src/openne/gf.py
        .. _GraRep: https://github.com/thunlp/OpenNE/blob/master/src/openne/grarep.py
        .. _HOPE: https://github.com/thunlp/OpenNE/blob/master/src/openne/hope.py
        .. _LaplacianEigenmaps: https://github.com/thunlp/OpenNE/blob/master/src/openne/lap.py
        .. _LINE: https://github.com/thunlp/OpenNE/blob/master/src/openne/line.py
        .. _LLE: https://github.com/thunlp/OpenNE/blob/master/src/openne/lle.py


    dimension: int
        the dimension of the label embedding vectors
    aggregation_function: 'add', 'multiply', 'average' or Callable
        the function used to aggregate label vectors for all labels assigned to each of the samples
    normalize_weights: boolean
        whether to normalize weights in the label graph by the number of samples or not
    param_dict
        parameters passed to the embedder, don't use the dimension and graph parameters, this class will set them at fit

    If you use this classifier please cite the relevant embedding method paper
    and the label network embedding for multi-label classification paper:

    .. code :: bibtex

        @article{zhang2007ml,
          title={ML-KNN: A lazy learning approach to multi-label learning},
          author={Zhang, Min-Ling and Zhou, Zhi-Hua},
          journal={Pattern recognition},
          volume={40},
          number={7},
          pages={2038--2048},
          year={2007},
          publisher={Elsevier}
        }

    Example code for using this embedder looks like this:

    .. code-block:: python

        from skmultilearn.embedding import OpenNetworkEmbedder, EmbeddingClassifier
        from sklearn.ensemble import RandomForestRegressor
        from skmultilearn.adapt import MLkNN
        from skmultilearn.cluster import LabelCooccurrenceGraphBuilder

        graph_builder = LabelCooccurrenceGraphBuilder(weighted=True, include_self_edges=False)
        openne_line_params = dict(batch_size=1000, negative_ratio=5)

        clf = EmbeddingClassifier(
            OpenNetworkEmbedder(graph_builder, 'LINE', 4, 'add', True, openne_line_params),
            RandomForestRegressor(n_estimators=10),
            MLkNN(k=5)
        )

        clf.fit(X_train, y_train)

        predictions = clf.predict(X_test)
    """

    _EMBEDDINGS = {
        "GraphFactorization": (GraphFactorization, "rep_size"),
        "GraRep": (GraRep, "dim"),
        "HOPE": (HOPE, "d"),
        "LaplacianEigenmaps": (LaplacianEigenmaps, "rep_size"),
        "LINE": (LINE, "rep_size"),
        "LLE": (LLE, "d"),
    }

    _AGGREGATION_FUNCTIONS = {
        "add": np.add.reduce,
        "multiply": np.multiply.reduce,
        "average": lambda x: np.average(x, axis=0),
    }

    def __init__(
        self,
        graph_builder,
        embedding,
        dimension,
        aggregation_function,
        normalize_weights,
        param_dict=None,
    ):
        if embedding not in self._EMBEDDINGS:
            raise ValueError(
                "Embedding must be one of {}".format(", ".join(self._EMBEDDINGS.keys()))
            )

        if aggregation_function in self._AGGREGATION_FUNCTIONS:
            self.aggregation_function = self._AGGREGATION_FUNCTIONS[
                aggregation_function
            ]
        elif callable(aggregation_function):
            self.aggregation_function = aggregation_function
        else:
            raise ValueError(
                "Aggregation function must be callable or one of {}".format(
                    ", ".join(self._AGGREGATION_FUNCTIONS.keys())
                )
            )

        self.embedding = embedding
        self.param_dict = param_dict if param_dict is not None else {}
        self.dimension = dimension
        self.graph_builder = graph_builder
        self.normalize_weights = normalize_weights

    def fit(self, X, y):
        self.fit_transform(X, y)

    def fit_transform(self, X, y):
        tf.compat.v1.reset_default_graph()
        self._init_openne_graph(y)
        embedding_class, dimension_key = self._EMBEDDINGS[self.embedding]
        param_dict = copy(self.param_dict)
        param_dict["graph"] = self.graph_
        param_dict[dimension_key] = self.dimension
        self.embeddings_ = embedding_class(**param_dict)
        return X, self._embedd_y(y)

    def _init_openne_graph(self, y):
        self.graph_ = Graph()
        self.graph_.G = nx.DiGraph()
        for (src, dst), w in self.graph_builder.transform(y).items():
            self.graph_.G.add_edge(src, dst)
            self.graph_.G.add_edge(dst, src)
            if self.normalize_weights:
                w = float(w) / y.shape[0]
            self.graph_.G[src][dst]["weight"] = w
            self.graph_.G[dst][src]["weight"] = w
        self.graph_.encode_node()

    def _embedd_y(self, y):
        empty_vector = np.zeros(shape=self.dimension)
        if sp.issparse(y):
            return np.array(
                [
                    self.aggregation_function(
                        [self.embeddings_.vectors[node] for node in row]
                    )
                    if len(row) > 0
                    else empty_vector
                    for row in _iterate_over_sparse_matrix(y)
                ]
            ).astype("float64")

        return np.array(
            [
                self.aggregation_function(
                    [
                        self.embeddings_.vectors[node]
                        for node, v in enumerate(row)
                        if v > 0
                    ]
                )
                if len(row) > 0
                else empty_vector
                for row in (y.A if isinstance(y, np.matrix) else y)
            ]
        ).astype("float64")


def _iterate_over_sparse_matrix(y):
    for r in range(y.shape[0]):
        yield y[r, :].indices



import numpy as np
from .base import MLClassifierBase
from scipy.sparse import csr_matrix


class MockClassifier(MLClassifierBase):
    """A stub classifier"""

    def __init__(self):
        super(MockClassifier, self).__init__()

    def fit(self, X, y):
        self.label_count = y.shape[1]
        return self

    def predict(self, X):
        return csr_matrix(np.ones(shape=(X.shape[0], self.label_count), dtype=int))



"""
The :mod:`skmultilearn.model_selection` module provides implementations multi-label stratification methods
useful for parameter estimation.

Currently the available modules include:

+-------------------------------------------------------------------+-------------------------------------------+
| Name                                                              | Description                               |
+===================================================================+===========================================+
| :class:`~skmultilearn.model_selection.iterative_stratification`   | Iterative stratification                  |
+-------------------------------------------------------------------+-------------------------------------------+
| :mod:`~skmultilearn.model_selection.measures`                     | Stratification quality measures package   |
+-------------------------------------------------------------------+-------------------------------------------+

"""

from .iterative_stratification import (
    IterativeStratification,
    iterative_train_test_split,
)
from .measures import (
    example_distribution,
    label_combination_distribution,
    percentage_of_label_combinations_without_evidence_per_fold,
    folds_without_evidence_for_at_least_one_label_combination,
    folds_label_combination_pairs_without_evidence,
)



from __future__ import absolute_import

import community
import networkx as nx
from networkx.algorithms.community import asyn_lpa_communities
import numpy as np

from .base import LabelGraphClustererBase
from .helpers import _membership_to_list_of_communities


class NetworkXLabelGraphClusterer(LabelGraphClustererBase):
    """Cluster label space with NetworkX community detection

    This clusterer constructs a NetworkX representation of the Label Graph generated by graph builder and detects
    communities in it using methods from the NetworkX library. Detected communities are converted to
    a label space clustering.

    Parameters
    ----------
    graph_builder: a GraphBuilderBase inherited transformer
        the graph builder to provide the adjacency matrix and weight map for the underlying graph
    method: string
        the community detection method to use, this clusterer supports the following community detection methods:

        +----------------------+--------------------------------------------------------------------------------+
        | Method name string   |                             Description                                        |
        +----------------------+--------------------------------------------------------------------------------+
        | louvain_             | Detecting communities with largest modularity using incremental greedy search  |
        +----------------------+--------------------------------------------------------------------------------+
        | label_propagation_   | Detecting communities from multiple async label propagation on the graph       |
        +----------------------+--------------------------------------------------------------------------------+

        .. _louvain: https://python-louvain.readthedocs.io/en/latest/
        .. _label_propagation: https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.community.label_propagation.asyn_lpa_communities.html


    Attributes
    ----------
    graph_ : networkx.Graph
        the networkx Graph object containing the graph representation of graph builder's adjacency matrix and weights
    weights_ : { 'weight' : list of values in edge order of graph edges }
        edge weights stored in a format recognizable by the networkx module

    References
    ----------
    If you use this clusterer please cite the igraph paper and the clustering paper:

    .. code :: latex

        @unknown{networkx,
            author = {Hagberg, Aric and Swart, Pieter and S Chult, Daniel},
            year = {2008},
            month = {01},
            title = {Exploring Network Structure, Dynamics, and Function Using NetworkX},
            booktitle = {Proceedings of the 7th Python in Science Conference}
        }

        @article{blondel2008fast,
          title={Fast unfolding of communities in large networks},
          author={Blondel, Vincent D and Guillaume, Jean-Loup and Lambiotte, Renaud and Lefebvre, Etienne},
          journal={Journal of statistical mechanics: theory and experiment},
          volume={2008},
          number={10},
          pages={P10008},
          year={2008},
          publisher={IOP Publishing}
        }


    Examples
    --------

    An example code for using this clusterer with a classifier looks like this:

    .. code-block:: python

        from sklearn.ensemble import RandomForestClassifier
        from skmultilearn.problem_transform import LabelPowerset
        from skmultilearn.cluster import NetworkXLabelGraphClusterer, LabelCooccurrenceGraphBuilder
        from skmultilearn.ensemble import LabelSpacePartitioningClassifier

        # construct base forest classifier
        base_classifier = RandomForestClassifier(n_estimators=1000)

        # construct a graph builder that will include
        # label relations weighted by how many times they
        # co-occurred in the data, without self-edges
        graph_builder = LabelCooccurrenceGraphBuilder(
            weighted = True,
            include_self_edges = False
        )

        # setup problem transformation approach with sparse matrices for random forest
        problem_transform_classifier = LabelPowerset(classifier=base_classifier,
            require_dense=[False, False])

        # setup the clusterer to use, we selected the modularity-based approach
        clusterer = NetworkXLabelGraphClusterer(graph_builder=graph_builder, method='louvain')

        # setup the ensemble metaclassifier
        classifier = LabelSpacePartitioningClassifier(problem_transform_classifier, clusterer)

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)

    For more use cases see `the label relations exploration guide <../labelrelations.ipynb>`_.

    """

    def __init__(self, graph_builder, method):
        """Initializes the clusterer

        Attributes
        ----------
        graph_builder: a GraphBuilderBase inherited transformer
                Class used to provide an underlying graph for NetworkX
        """
        super(NetworkXLabelGraphClusterer, self).__init__(graph_builder)
        self.method = method

    def fit_predict(self, X, y):
        """Performs clustering on y and returns list of label lists

        Builds a label graph using the provided graph builder's `transform` method
        on `y` and then detects communities using the selected `method`.

        Sets :code:`self.weights_` and :code:`self.graph_`.

        Parameters
        ----------
        X : None
            currently unused, left for scikit compatibility
        y : scipy.sparse
            label space of shape :code:`(n_samples, n_labels)`

        Returns
        -------
        arrray of arrays of label indexes (numpy.ndarray)
            label space division, each sublist represents labels that are in that community
        """
        edge_map = self.graph_builder.transform(y)

        if self.graph_builder.is_weighted:
            self.weights_ = dict(weight=list(edge_map.values()))
        else:
            self.weights_ = dict(weight=None)

        self.graph_ = nx.Graph()
        for n in range(y.shape[1]):
            self.graph_.add_node(n)

        for e, w in edge_map.items():
            self.graph_.add_edge(e[0], e[1], weight=w)

        if self.method == "louvain":
            partition_dict = community.best_partition(self.graph_)
            memberships = [partition_dict[i] for i in range(y.shape[1])]

            return np.array(
                _membership_to_list_of_communities(memberships, 1 + max(memberships))
            )
        else:
            return np.array(
                [list(i) for i in asyn_lpa_communities(self.graph_, "weight")]
            )



import numpy as np
from .base import MLClassifierBase
from ..utils import matrix_creation_function_for_format
from scipy.sparse import issparse, csr_matrix


class ProblemTransformationBase(MLClassifierBase):
    """Base class providing common functions for multi-label classifiers
    that follow the problem transformation approach.

    Problem transformation is the approach in which the
    original multi-label classification problem is transformed into one
    or more single-label problems, which are then solved by single-class
    or multi-class classifiers.

    Scikit-multilearn provides a number of such methods:

    - :class:`BinaryRelevance` - performs a single-label single-class classification for each label and sums the results :class:`BinaryRelevance`
    - :class:`ClassifierChains` - performs a single-label single-class classification for each label and sums the results :class:`ClassifierChain`
    - :class:`LabelPowerset` - performs a single-label single-class classification for each label and sums the results :class:`LabelPowerset`

    Parameters
    ----------
    classifier : scikit classifier type
        The base classifier that will be used in a class, will be automagically put under self.classifier for future access.
    require_dense : boolean (default is False)
        Whether the base classifier requires input as dense arrays.
    """

    def __init__(self, classifier=None, require_dense=None):
        super(ProblemTransformationBase, self).__init__()

        self.copyable_attrs = ["classifier", "require_dense"]

        self.classifier = classifier
        if require_dense is not None:
            if isinstance(require_dense, bool):
                self.require_dense = [require_dense, require_dense]
            else:
                assert (
                    len(require_dense) == 2
                    and isinstance(require_dense[0], bool)
                    and isinstance(require_dense[1], bool)
                )
                self.require_dense = require_dense

        else:
            if isinstance(self.classifier, MLClassifierBase):
                self.require_dense = [False, False]
            else:
                self.require_dense = [True, True]

    def _ensure_multi_label_from_single_class(self, matrix, matrix_format="csr"):
        """Transform single class outputs to a 2D sparse matrix

        Parameters
        ----------
        matrix : array-like
            input matrix to be checked
        matrix_format : str (default is csr)
            the matrix format to validate with

        Returns
        -------
        scipy.sparse
            a 2-dimensional sparse matrix
        """
        is_2d = None
        dim_1 = None
        dim_2 = None

        # check if array like of array likes
        if isinstance(matrix, (list, tuple, np.ndarray)):
            if isinstance(matrix[0], (list, tuple, np.ndarray)):
                is_2d = True
                dim_1 = len(matrix)
                dim_2 = len(matrix[0])
            # 1d list or array
            else:
                is_2d = False
                # shape is n_samples of 1 class assignment
                dim_1 = len(matrix)
                dim_2 = 1

        # not an array but 2D, probably a matrix
        elif matrix.ndim == 2:
            is_2d = True
            dim_1 = matrix.shape[0]
            dim_2 = matrix.shape[1]

        # what is it?
        else:
            raise ValueError("Matrix dimensions too large (>2) or other value error")

        new_matrix = None
        if is_2d:
            if issparse(matrix):
                new_matrix = matrix
            else:
                new_matrix = matrix_creation_function_for_format(matrix_format)(
                    matrix, shape=(dim_1, dim_2)
                )
        else:
            new_matrix = matrix_creation_function_for_format(matrix_format)(matrix).T

        assert new_matrix.shape == (dim_1, dim_2)
        return new_matrix



"""
The :mod:`skmultilearn.adapt` module implements algorithm
adaptation approaches to multi-label classification.

Algorithm adaptation methods for multi-label classification 
concentrate on adapting single-label classification algorithms
to the multi-label case usually by changes in cost/decision 
functions.

Currently the following algorithm adaptation classification schemes are available in scikit-multilearn:

+-----------------------------------------------+-----------------------------------------------------------+
| Classifier                                    | Description                                               |
+===============================================+===========================================================+
| :class:`~skmultilearn.adapt.BRkNNaClassifier` | a Binary Relevance kNN classifier that assigns a label    |
|                                               | if at least half of the neighbors are also classified     |
|                                               | with the label                                            |
+-----------------------------------------------+-----------------------------------------------------------+
| :class:`~skmultilearn.adapt.BRkNNbClassifier` | a Binary Relevance kNN classifier that assigns top m      |
|                                               | labels of neighbors with m - average number of labels     |
|                                               | assigned to neighbors                                     |
+-----------------------------------------------+-----------------------------------------------------------+
| :class:`~skmultilearn.adapt.MLkNN`            | a multi-label adapted kNN classifier with bayesian prior  |
|                                               | corrections                                               |
+-----------------------------------------------+-----------------------------------------------------------+
| :class:`~skmultilearn.adapt.MLARAM`           | a multi-Label Hierarchical ARAM Neural Network            |
+-----------------------------------------------+-----------------------------------------------------------+
| :class:`~skmultilearn.adapt.MLTSVM`           | twin multi-Label Support Vector Machines                  |
+-----------------------------------------------+-----------------------------------------------------------+

"""

from .brknn import BRkNNaClassifier, BRkNNbClassifier
from .mlknn import MLkNN
from .mlaram import MLARAM
from .mltsvm import MLTSVM

__all__ = ["BRkNNaClassifier", "BRkNNbClassifier", "MLkNN", "MLARAM", "MLTSVM"]



from builtins import range
from ..base import MLClassifierBase
from ..utils import get_matrix_in_format
from sklearn.neighbors import NearestNeighbors
import scipy.sparse as sparse
import numpy as np


class _BinaryRelevanceKNN(MLClassifierBase):
    """Binary Relevance adapted kNN Multi-Label Classifier base class."""

    def __init__(self, k=10):
        super(_BinaryRelevanceKNN, self).__init__()
        self.k = k  # Number of neighbours
        self.copyable_attrs = ["k"]

    def fit(self, X, y):
        """Fit classifier with training data

        Internally this method uses a sparse CSC representation for y
        (:class:`scipy.sparse.csc_matrix`).

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse
            input features, can be a dense or sparse matrix of size
            :code:`(n_samples, n_features)`
        y : numpy.ndaarray or scipy.sparse {0,1}
            binary indicator matrix with label assignments.

        Returns
        -------
        self
            fitted instance of self
        """
        self.train_labelspace = get_matrix_in_format(y, "csc")
        self._n_samples = self.train_labelspace.shape[0]
        self._n_labels = self.train_labelspace.shape[1]
        self.knn_ = NearestNeighbors(n_neighbors=self.k).fit(X)
        return self

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse of int
            binary indicator matrix with label assignments with shape
            :code:`(n_samples, n_labels)`
        """
        self.neighbors_ = self.knn_.kneighbors(X, self.k, return_distance=False)
        self.confidences_ = np.vstack(
            [
                self.train_labelspace[n, :].tocsc().sum(axis=0) / self.k
                for n in self.neighbors_
            ]
        )
        return self._predict_variant(X)


class BRkNNaClassifier(_BinaryRelevanceKNN):
    """Binary Relevance multi-label classifier based on k-Nearest Neighbors method.

    This version of the classifier assigns the labels that are assigned
    to at least half of the neighbors.

    Parameters
    ----------
    k : int
        number of neighbours


    Attributes
    ----------
    knn_ : an instance of sklearn.NearestNeighbors
        the nearest neighbors single-label classifier used underneath
    neighbors_ : array of arrays of int, shape = (n_samples, k)
        k neighbors of each sample

    confidences_ : matrix of int, shape = (n_samples, n_labels)
        label assignment confidences


    References
    ----------

    If you use this method please cite the relevant paper:

    .. code :: bibtex

         @inproceedings{EleftheriosSpyromitros2008,
            author = {Eleftherios Spyromitros, Grigorios Tsoumakas, Ioannis Vlahavas},
            booktitle = {Proc. 5th Hellenic Conference on Artificial Intelligence (SETN 2008)},
            title = {An Empirical Study of Lazy Multilabel Classification Algorithms},
            year = {2008},
            location = {Syros, Greece}
         }

    Examples
    --------

    Here's a very simple example of using BRkNNaClassifier with a fixed number of neighbors:

    .. code :: python

        from skmultilearn.adapt import BRkNNaClassifier

        classifier = BRkNNaClassifier(k=3)

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)


    You can also use :class:`~sklearn.model_selection.GridSearchCV` to find an optimal set of parameters:

    .. code :: python

        from skmultilearn.adapt import BRkNNaClassifier
        from sklearn.model_selection import GridSearchCV

        parameters = {'k': range(1,3)}
        score = 'f1_macro'

        clf = GridSearchCV(BRkNNaClassifier(), parameters, scoring=score)
        clf.fit(X, y)

    """

    def _predict_variant(self, X):
        # TODO: find out if moving the sparsity to compute confidences_ boots speed
        return sparse.csr_matrix(np.rint(self.confidences_), dtype="i8")


class BRkNNbClassifier(_BinaryRelevanceKNN):
    """Binary Relevance multi-label classifier based on k-Nearest Neighbors method.

    This version of the classifier assigns the most popular m labels of
    the neighbors, where m is the  average number of labels assigned to
    the object's neighbors.

    Parameters
    ----------
    k : int
        number of neighbours

    Attributes
    ----------
    knn_ : an instance of sklearn.NearestNeighbors
        the nearest neighbors single-label classifier used underneath
    neighbors_ : array of arrays of int, shape = (n_samples, k)
        k neighbors of each sample

    confidences_ : matrix of int, shape = (n_samples, n_labels)
        label assignment confidences


    References
    ----------

    If you use this method please cite the relevant paper:

    .. code :: bibtex

         @inproceedings{EleftheriosSpyromitros2008,
            author = {Eleftherios Spyromitros, Grigorios Tsoumakas, Ioannis Vlahavas},
            booktitle = {Proc. 5th Hellenic Conference on Artificial Intelligence (SETN 2008)},
            title = {An Empirical Study of Lazy Multilabel Classification Algorithms},
            year = {2008},
            location = {Syros, Greece}
         }

    Examples
    --------

    Here's a very simple example of using BRkNNbClassifier with a fixed number of neighbors:

    .. code :: python

        from skmultilearn.adapt import BRkNNbClassifier

        classifier = BRkNNbClassifier(k=3)

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)


    You can also use :class:`~sklearn.model_selection.GridSearchCV` to find an optimal set of parameters:

    .. code :: python

        from skmultilearn.adapt import BRkNNbClassifier
        from sklearn.model_selection import GridSearchCV

        parameters = {'k': range(1,3)}
        score = 'f1-macro

        clf = GridSearchCV(BRkNNbClassifier(), parameters, scoring=score)
        clf.fit(X, y)

    """

    def _predict_variant(self, X):
        avg_labels = [
            int(np.average(self.train_labelspace[n, :].sum(axis=1)).round())
            for n in self.neighbors_
        ]

        prediction = sparse.lil_matrix((X.shape[0], self._n_labels), dtype="i8")
        top_labels = np.argpartition(
            self.confidences_, kth=min(avg_labels + [len(self.confidences_[0])]), axis=1
        ).tolist()

        for i in range(X.shape[0]):
            for j in top_labels[i][-avg_labels[i] :]:
                prediction[i, j] += 1

        return prediction



from __future__ import absolute_import

from sklearn.base import BaseEstimator


class SKLearnEmbedder(BaseEstimator):
    """Embed the label space using a scikit-compatible matrix-based embedder

    Parameters
    ----------
    embedder : sklearn.base.BaseEstimator
        a clonable instance of a scikit-compatible embedder, will be automatically
        put under :code:`self.embedder`, see .
    pass_input_space : bool (default is False)
        whether to take :code:`X` into consideration upon clustering,
        use only if you know that the embedder can handle two
        parameters for clustering, will be automatically
        put under :code:`self.pass_input_space`.


    Example code for using this embedder looks like this:

    .. code-block:: python

        from skmultilearn.embedding import SKLearnEmbedder, EmbeddingClassifier
        from sklearn.manifold import SpectralEmbedding
        from sklearn.ensemble import RandomForestRegressor
        from skmultilearn.adapt import MLkNN

        clf = EmbeddingClassifier(
            SKLearnEmbedder(SpectralEmbedding(n_components = 10)),
            RandomForestRegressor(n_estimators=10),
            MLkNN(k=5)
        )

        clf.fit(X_train, y_train)

        predictions = clf.predict(X_test)
    """

    def __init__(self, embedder=None, pass_input_space=False):
        super(BaseEstimator, self).__init__()

        self.embedder = embedder
        self.pass_input_space = pass_input_space

    def fit(self, X, y):
        """Fits the embedder to data

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self
        """

        self.embedder.fit(X, y)

    def fit_transform(self, X, y):
        """Fit the embedder and transform the output space

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        X, y_embedded
            results of the embedding, input and output space
        """

        if self.pass_input_space:
            result = self.embedder.fit_transform(X, y)
        else:
            result = self.embedder.fit_transform(y)

        return X, result



"""
Multi-label classification module for Python
============================================

Scikit-multilearn is a BSD-licensed library for multi-label classification that is
built on top of the well-known scikit-learn ecosystem.

"""



# copyright @Fernando Benites
from builtins import object
from builtins import range

import numpy
import numpy.core.umath as umath
import scipy.sparse
from scipy.sparse import issparse

from ..base import MLClassifierBase


class Neuron(object):
    """An implementation of a neuron for MLARAM

    Parameters
    ----------
    vc : array
        neuron's assigned vector
    label : int
        label number
    """

    def __init__(self, vc, label):
        # vector must be in complement form
        self.vc = vc
        self.label = label


def _get_label_combination_representation(label_assignment_binary_indicator_list):
    return label_assignment_binary_indicator_list.nonzero()[0].tobytes().decode('utf-8')


def _get_label_vector(y, i):
    if issparse(y):
        return numpy.squeeze(numpy.asarray(y[i].todense()))
    return y[i]


def _concatenate_with_negation(row):
    ones = numpy.ones(row.shape)
    if issparse(row):
        return scipy.sparse.hstack((row, ones - row))
    else:
        # concatenate and merge sublists in the row if it is matrix
        return numpy.concatenate((row, ones - row), int(len(row.shape) != 1))


def _normalize_input_space(X):
    x_max = X.max()
    x_min = X.min()
    if x_max < 0 or x_max > 1 or x_min < 0 or x_min > 1:
        return numpy.multiply(X - x_min, 1 / (x_max - x_min))
    return X


class MLARAM(MLClassifierBase):
    """HARAM: A Hierarchical ARAM Neural Network for Large-Scale Text Classification

    This method aims at increasing the classification speed by adding an
    extra ART layer for clustering learned prototypes into large clusters.
    In this case the activation of all prototypes can be replaced by the
    activation of a small fraction of them, leading to a significant
    reduction of the classification time.

    Parameters
    ----------
    vigilance : float (default is 0.9)
        parameter for adaptive resonance theory networks, controls how
        large a hyperbox can be, 1 it is small (no compression), 0
        should assume all range. Normally set between 0.8 and 0.999,
        it is dataset dependent. It is responsible for the creation
        of the prototypes, therefore training of the network.
    threshold : float (default is 0.02)
        controls how many prototypes participate by the prediction,
        can be changed for the testing phase.
    neurons : list
        the neurons in the network


    References
    ----------

    Published work available `here`_.

    .. _here: http://dx.doi.org/10.1109/ICDMW.2015.14

    .. code :: bibtex

        @INPROCEEDINGS{7395756,
            author={F. Benites and E. Sapozhnikova},
            booktitle={2015 IEEE International Conference on Data Mining Workshop (ICDMW)},
            title={HARAM: A Hierarchical ARAM Neural Network for Large-Scale Text Classification},
            year={2015},
            volume={},
            number={},
            pages={847-854},
            doi={10.1109/ICDMW.2015.14},
            ISSN={2375-9259},
            month={Nov},
        }

    Examples
    --------

    Here's an example code with a 5% threshold and vigilance of 0.95:

    .. code :: python

        from skmultilearn.adapt import MLARAM

        classifier = MLARAM(threshold=0.05, vigilance=0.95)
        classifier.fit(X_train, y_train)
        prediction = classifier.predict(X_test)


    """

    def __init__(self, vigilance=0.9, threshold=0.02, neurons=None):
        super(MLARAM, self).__init__()

        if neurons is not None:
            self.neurons = neurons
        else:
            self.neurons = []
        self.vigilance = vigilance
        self.threshold = threshold

        self.copyable_attrs += ["neurons", "vigilance", "threshold"]

    def reset(self):
        """Resets the labels and neurons"""
        self._labels = []
        self.neurons = []

    def fit(self, X, y):
        """Fit classifier with training data

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse
            input features, can be a dense or sparse matrix of size
            :code:`(n_samples, n_features)`
        y : numpy.ndarray or scipy.sparse {0,1}
            binary indicator matrix with label assignments.

        Returns
        -------
        skmultilearn.MLARAMfast.MLARAM
            fitted instance of self
        """

        self._labels = []
        self._allneu = ""
        self._online = 1
        self._alpha = 0.0000000000001

        is_sparse_x = issparse(X)

        label_combination_to_class_map = {}
        # FIXME: we should support dense matrices natively
        if isinstance(X, numpy.matrix):
            X = numpy.asarray(X)
        if isinstance(y, numpy.matrix):
            y = numpy.asarray(y)
        is_more_dimensional = int(len(X[0].shape) != 1)
        X = _normalize_input_space(X)

        y_0 = _get_label_vector(y, 0)

        if len(self.neurons) == 0:
            neuron_vc = _concatenate_with_negation(X[0])
            self.neurons.append(Neuron(neuron_vc, y_0))
            start_index = 1
            label_combination_to_class_map[
                _get_label_combination_representation(y_0)
            ] = [0]
        else:
            start_index = 0

        # denotes the class enumerator for label combinations
        last_used_label_combination_class_id = 0

        for row_no, input_vector in enumerate(X[start_index:], start_index):
            label_assignment_vector = _get_label_vector(y, row_no)

            fc = _concatenate_with_negation(input_vector)
            activationn = [0] * len(self.neurons)
            activationi = [0] * len(self.neurons)
            label_combination = _get_label_combination_representation(
                label_assignment_vector
            )

            if label_combination in label_combination_to_class_map:
                fcs = fc.sum()
                for class_number in label_combination_to_class_map[label_combination]:
                    if issparse(self.neurons[class_number].vc):
                        minnfs = self.neurons[class_number].vc.minimum(fc).sum()
                    else:
                        minnfs = umath.minimum(self.neurons[class_number].vc, fc).sum()

                    activationi[class_number] = minnfs / fcs
                    activationn[class_number] = (
                        minnfs / self.neurons[class_number].vc.sum()
                    )

            if numpy.max(activationn) == 0:
                last_used_label_combination_class_id += 1
                self.neurons.append(Neuron(fc, label_assignment_vector))
                label_combination_to_class_map.setdefault(label_combination, []).append(
                    len(self.neurons) - 1
                )

                continue

            inds = numpy.argsort(activationn)
            indc = numpy.where(numpy.array(activationi)[inds[::-1]] > self.vigilance)[0]

            if indc.shape[0] == 0:
                self.neurons.append(Neuron(fc, label_assignment_vector))
                label_combination_to_class_map.setdefault(label_combination, []).append(
                    len(self.neurons) - 1
                )
                continue

            winner = inds[::-1][indc[0]]
            if issparse(self.neurons[winner].vc):
                self.neurons[winner].vc = self.neurons[winner].vc.minimum(fc)
            else:
                self.neurons[winner].vc = umath.minimum(self.neurons[winner].vc, fc)

            # 1 if winner neuron won a given label 0 if not
            labels_won_indicator = numpy.zeros(y_0.shape, dtype=y_0.dtype)
            labels_won_indicator[label_assignment_vector.nonzero()] = 1
            self.neurons[winner].label += labels_won_indicator

        return self

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse of int
            binary indicator matrix with label assignments with shape
            :code:`(n_samples, n_labels)`
        """

        result = []
        # FIXME: we should support dense matrices natively
        if isinstance(X, numpy.matrix):
            X = numpy.asarray(X)
        ranks = self.predict_proba(X)
        for rank in ranks:
            sorted_rank_arg = numpy.argsort(-rank)
            diffs = -numpy.diff([rank[k] for k in sorted_rank_arg])

            indcutt = numpy.where(diffs == diffs.max())[0]
            if len(indcutt.shape) == 1:
                indcut = indcutt[0] + 1
            else:
                indcut = indcutt[0, -1] + 1
            label = numpy.zeros(rank.shape)

            label[sorted_rank_arg[0:indcut]] = 1

            result.append(label)

        return numpy.array(result)

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        array of arrays of float
            matrix with label assignment probabilities of shape
            :code:`(n_samples, n_labels)`
        """
        # FIXME: we should support dense matrices natively
        if isinstance(X, numpy.matrix):
            X = numpy.asarray(X)
        if issparse(X):
            if X.getnnz() == 0:
                return
        elif len(X) == 0:
            return

        is_matrix = int(len(X[0].shape) != 1)
        X = _normalize_input_space(X)

        all_ranks = []
        neuron_vectors = [n1.vc for n1 in self.neurons]
        if any(map(issparse, neuron_vectors)):
            all_neurons = scipy.sparse.vstack(neuron_vectors)
            # can't add a constant to a sparse matrix in scipy
            all_neurons_sum = all_neurons.sum(1).A
        else:
            all_neurons = numpy.vstack(neuron_vectors)
            all_neurons_sum = all_neurons.sum(1)

        all_neurons_sum += self._alpha

        for row_number, input_vector in enumerate(X):
            fc = _concatenate_with_negation(input_vector)

            if issparse(fc):
                activity = (
                    (fc.minimum(all_neurons).sum(1) / all_neurons_sum)
                    .squeeze()
                    .tolist()
                )
            else:
                activity = (
                    (umath.minimum(fc, all_neurons).sum(1) / all_neurons_sum)
                    .squeeze()
                    .tolist()
                )

            if is_matrix:
                activity = activity[0]

            # be very fast
            sorted_activity = numpy.argsort(activity)[::-1]
            winner = sorted_activity[0]
            activity_difference = activity[winner] - activity[sorted_activity[-1]]
            largest_activity = 1
            par_t = self.threshold

            for i in range(1, len(self.neurons)):
                activity_change = (
                    activity[winner] - activity[sorted_activity[i]]
                ) / activity[winner]
                if activity_change > par_t * activity_difference:
                    break

                largest_activity += 1

            rbsum = sum([activity[k] for k in sorted_activity[0:largest_activity]])
            rank = activity[winner] * self.neurons[winner].label
            activated = []
            activity_among_activated = []
            activated.append(winner)
            activity_among_activated.append(activity[winner])

            for i in range(1, largest_activity):
                rank += (
                    activity[sorted_activity[i]]
                    * self.neurons[sorted_activity[i]].label
                )
                activated.append(sorted_activity[i])
                activity_among_activated.append(activity[sorted_activity[i]])

            rank /= rbsum
            all_ranks.append(rank)

        return numpy.array(all_ranks)



import arff
import bz2
import pickle
from scipy import sparse
import hashlib

import os
import requests
import shutil
from collections import defaultdict


def get_data_home(data_home=None, subdirectory=""):
    """Return the path of the scikit-multilearn data dir.

    This folder is used by some large dataset loaders to avoid
    downloading the data several times.

    By default the :code:`data_home` is set to a folder named
    :code:`'scikit_ml_learn_data'` in the user home folder.

    Alternatively, it can be set by the :code:`'SCIKIT_ML_LEARN_DATA'`
    environment variable or programmatically by giving an explicit
    folder path. The :code:`'~'` symbol is expanded to the user home
    folder.

    If the folder does not already exist, it is automatically created.

    Parameters
    ----------
    data_home : str (default is None)
        the path to the directory in which scikit-multilearn data sets
        should be stored, if None the path is generated as stated above

    subdirectory : str, default ''
        return path subdirectory under data_home if data_home passed or under default if not passed

    Returns
    --------
    str
        the path to the data home
    """
    if data_home is None:
        if len(subdirectory) > 0:
            data_home = os.environ.get(
                "SCIKIT_ML_LEARN_DATA",
                os.path.join("~", "scikit_ml_learn_data", subdirectory),
            )
        else:
            data_home = os.environ.get(
                "SCIKIT_ML_LEARN_DATA", os.path.join("~", "scikit_ml_learn_data")
            )
    data_home = os.path.expanduser(data_home)
    if not os.path.exists(data_home):
        os.makedirs(data_home)
    return data_home


def clear_data_home(data_home=None):
    """Delete all the content of the data home cache.

    Parameters
    ----------
    data_home : str (default is None)
        the path to the directory in which scikit-multilearn data sets
        should be stored.
    """
    data_home = get_data_home(data_home)
    shutil.rmtree(data_home)


def _get_download_base_url():
    """Returns base URL for data sets."""

    return "http://scikit.ml/datasets/"


def available_data_sets():
    """Lists available data sets and their variants

    Returns
    -------
    dict[(set_name, variant_name)] -> [md5, file_name]
        available datasets and their variants with the key pertaining
        to the :code:`(set_name, variant_name)` and values include md5 and file name on server
    """
    r = requests.get(_get_download_base_url() + "data.list")
    if r.status_code != 200:
        r.raise_for_status()
    else:
        raw_data_list = r.text

        variant_information = defaultdict(list)
        for row in raw_data_list.split("\n"):
            md5, file_name = row.split(";")
            set_name, variant = file_name.split(".")[0].split("-")
            if (set_name, variant) in variant_information:
                raise Exception(
                    "Data file broken, files doubled, please file bug report."
                )
            variant_information[(set_name, variant)] = [md5, file_name]
        return variant_information


def download_dataset(set_name, variant, data_home=None):
    """Downloads a data set

    Parameters
    ----------
    set_name : str
        name of set from :func:`available_data_sets`
    variant : str
        variant of the data set from :func:`available_data_sets`
    data_home : default None, str
        custom base folder for data, if None, default is used

    Returns
    -------
    str
        path to the downloaded data set file on disk
    """

    data_sets = available_data_sets()
    if (set_name, variant) not in data_sets:
        raise ValueError(
            "The set {} in variant {} does not exist on server.".format(
                set_name, variant
            )
        )

    md5, name = data_sets[set_name, variant]

    if data_home is None:
        target_name = os.path.join(get_data_home(), name)
    else:
        target_name = os.path.join(data_home, name)

    if os.path.exists(target_name):
        if md5 == _get_md5(target_name):
            print("{}:{} - exists, not redownloading".format(set_name, variant))
            return target_name
        else:
            print(
                "{}:{} - exists, but MD5 sum mismatch - redownloading".format(
                    set_name, variant
                )
            )
    else:
        print("{}:{} - does not exists downloading".format(set_name, variant))

    # not found or broken md5
    _download_single_file(name, target_name)
    found_md5 = _get_md5(target_name)
    if md5 != found_md5:
        raise Exception(
            "{}: MD5 mismatch {} vs {} - possible download error".format(
                name, md5, found_md5
            )
        )

    print("Downloaded {}-{}".format(set_name, variant))

    return target_name


def load_dataset(set_name, variant, data_home=None):
    """Loads a selected variant of the given data set

    Parameters
    ----------
    set_name : str
        name of set from :func:`available_data_sets`
    variant : str
        variant of the data set
    data_home : default None, str
        custom base folder for data, if None, default is used

    Returns
    --------
    dict
        the loaded multilabel data set variant in the scikit-multilearn
        format, see data_sets
    """

    path = download_dataset(set_name, variant, data_home)
    if path is not None:
        return load_dataset_dump(path)

    return None


def load_from_arff(
    filename,
    label_count,
    label_location="end",
    input_feature_type="float",
    encode_nominal=True,
    load_sparse=False,
    return_attribute_definitions=False,
):
    """Method for loading ARFF files as numpy array

    Parameters
    ----------
    filename : str
        path to ARFF file
    labelcount: integer
        number of labels in the ARFF file
    endian: str {"big", "little"} (default is "big")
        whether the ARFF file contains labels at the beginning of the
        attributes list ("start", MEKA format)
        or at the end ("end", MULAN format)
    input_feature_type: numpy.type as string (default is "float")
        the desire type of the contents of the return 'X' array-likes,
        default 'i8', should be a numpy type,
        see http://docs.scipy.org/doc/numpy/user/basics.types.html
    encode_nominal: bool (default is True)
        whether convert categorical data into numeric factors - required
        for some scikit classifiers that can't handle non-numeric
        input features.
    load_sparse: boolean (default is False)
        whether to read arff file as a sparse file format, liac-arff
        breaks if sparse reading is enabled for non-sparse ARFFs.
    return_attribute_definitions: boolean (default is False)
        whether to return the definitions for each attribute in the
        dataset

    Returns
    -------
    X : :mod:`scipy.sparse.lil_matrix` of `input_feature_type`, shape=(n_samples, n_features)
        input feature matrix
    y : :mod:`scipy.sparse.lil_matrix` of `{0, 1}`, shape=(n_samples, n_labels)
        binary indicator matrix with label assignments
    names of attributes : List[str]
        list of attribute names from ARFF file
    """

    if not load_sparse:
        arff_frame = arff.load(
            open(filename, "r"), encode_nominal=encode_nominal, return_type=arff.DENSE
        )
        matrix = sparse.csr_matrix(arff_frame["data"], dtype=input_feature_type)
    else:
        arff_frame = arff.load(
            open(filename, "r"), encode_nominal=encode_nominal, return_type=arff.COO
        )
        data = arff_frame["data"][0]
        row = arff_frame["data"][1]
        col = arff_frame["data"][2]
        matrix = sparse.coo_matrix(
            (data, (row, col)), shape=(max(row) + 1, max(col) + 1)
        )

    if label_location == "start":
        X, y = (
            matrix.tocsc()[:, label_count:].tolil(),
            matrix.tocsc()[:, :label_count].astype(int).tolil(),
        )
        feature_names = arff_frame["attributes"][label_count:]
        label_names = arff_frame["attributes"][:label_count]
    elif label_location == "end":
        X, y = (
            matrix.tocsc()[:, :-label_count].tolil(),
            matrix.tocsc()[:, -label_count:].astype(int).tolil(),
        )
        feature_names = arff_frame["attributes"][:-label_count]
        label_names = arff_frame["attributes"][-label_count:]
    else:
        # unknown endian
        return None

    if return_attribute_definitions:
        return X, y, feature_names, label_names
    else:
        return X, y


def save_to_arff(X, y, label_location="end", save_sparse=True, filename=None):
    """Method for dumping data to ARFF files

    Parameters
    ----------
    X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
        input feature matrix
    y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
        binary indicator matrix with label assignments
    label_location: string {"start", "end"} (default is "end")
        whether the ARFF file will contain labels at the beginning of the
        attributes list ("start", MEKA format)
        or at the end ("end", MULAN format)
    save_sparse: boolean
        Whether to save in ARFF's sparse dictionary-like format instead of listing all
        zeroes within file, very useful in multi-label classification.
    filename : str or None
        Path to ARFF file, if None, the ARFF representation is returned as string
    Returns
    -------
    str or None
        the ARFF dump string, if filename is None
    """
    X = X.todok()
    y = y.todok()

    x_prefix = 0
    y_prefix = 0

    x_attributes = [("X{}".format(i), "NUMERIC") for i in range(X.shape[1])]
    y_attributes = [("y{}".format(i), [str(0), str(1)]) for i in range(y.shape[1])]

    if label_location == "end":
        y_prefix = X.shape[1]
        relation_sign = -1
        attributes = x_attributes + y_attributes

    elif label_location == "start":
        x_prefix = y.shape[1]
        relation_sign = 1
        attributes = y_attributes + x_attributes

    else:
        raise ValueError("Label location not in {start, end}")

    if save_sparse:
        data = [{} for r in range(X.shape[0])]
    else:
        data = [[0 for c in range(X.shape[1] + y.shape[1])] for r in range(X.shape[0])]

    for keys, value in list(X.items()):
        data[keys[0]][x_prefix + keys[1]] = value

    for keys, value in list(y.items()):
        data[keys[0]][y_prefix + keys[1]] = value

    dataset = {
        "description": "traindata",
        "relation": "traindata: -C {}".format(y.shape[1] * relation_sign),
        "attributes": attributes,
        "data": data,
    }

    arff_data = arff.dumps(dataset)

    if filename is None:
        return arff_data

    with open(filename, "w") as fp:
        fp.write(arff_data)


def save_dataset_dump(input_space, labels, feature_names, label_names, filename=None):
    """Saves a compressed data set dump

    Parameters
    ----------
    input_space: array-like of array-likes
        Input space array-like of input feature vectors
    labels: array-like of binary label vectors
        Array-like of labels assigned to each input vector, as a binary
        indicator vector (i.e. if 5th position has value 1
        then the input vector has label no. 5)
    feature_names: array-like,optional
        names of features
    label_names: array-like, optional
        names of labels
    filename : str, optional
        Path to dump file, if without .bz2, the .bz2 extension will be
        appended.
    """
    data = {
        "X": input_space,
        "y": labels,
        "features": feature_names,
        "labels": label_names,
    }

    if filename is not None:
        if filename[-4:] != ".bz2":
            filename += ".bz2"

        with bz2.BZ2File(filename, "wb") as file_handle:
            pickle.dump(data, file_handle)
    else:
        return data


def load_dataset_dump(filename):
    """Loads a compressed data set dump

    Parameters
    ----------
    filename : str
        path to dump file, if without .bz2 ending, the .bz2 extension will be appended.

    Returns
    -------
    X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
        input feature matrix
    y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
        binary indicator matrix with label assignments
    names of attributes: List[str]
        list of attribute names for `X` columns
    names of labels: List[str]
        list of label names for `y` columns
    """

    if not os.path.exists(filename):
        raise IOError(
            "File {} does not exist, use load_dataset to download file".format(filename)
        )

    if filename[-4:] != ".bz2":
        filename += ".bz2"

    with bz2.BZ2File(filename, "r") as file_handle:
        data = pickle.load(file_handle)

    return data["X"], data["y"], data["features"], data["labels"]


def _download_single_file(data_file_name, target_file_name, base_url=None):
    base_url = base_url or _get_download_base_url()
    r = requests.get(base_url + data_file_name, stream=True)
    if r.status_code == 200:
        with open(target_file_name, "wb") as f:
            r.raw.decode_content = True
            shutil.copyfileobj(r.raw, f)
    else:
        r.raise_for_status()


def _get_md5(file_name):
    hash_md5 = hashlib.md5()
    with open(file_name, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



import unittest
import sys, platform

from skmultilearn.adapt import MLkNN
from skmultilearn.cluster import LabelCooccurrenceGraphBuilder
from skmultilearn.embedding import CLEMS, SKLearnEmbedder, EmbeddingClassifier
from skmultilearn.tests.classifier_basetest import ClassifierBaseTest
from sklearn.linear_model import LinearRegression
from sklearn.manifold import SpectralEmbedding
from copy import copy
import sklearn.metrics as metrics

if not (sys.version_info[0] == 2 or platform.architecture()[0] == "32bit"):
    from skmultilearn.embedding import OpenNetworkEmbedder


class EmbeddingTest(ClassifierBaseTest):
    TEST_NEIGHBORS = 3

    def classifiers(self):
        graph_builder = LabelCooccurrenceGraphBuilder(
            weighted=True, include_self_edges=False
        )

        param_dicts = {
            "GraphFactorization": dict(epoch=1),
            "GraRep": dict(Kstep=2),
            "HOPE": dict(),
            "LaplacianEigenmaps": dict(),
            "LINE": dict(epoch=1, order=1),
            "LLE": dict(),
        }

        if not (sys.version_info[0] == 2 or platform.architecture()[0] == "32bit"):
            for embedding in OpenNetworkEmbedder._EMBEDDINGS:
                if embedding == "LLE":
                    dimension = 3
                else:
                    dimension = 4

                yield EmbeddingClassifier(
                    OpenNetworkEmbedder(
                        copy(graph_builder),
                        embedding,
                        dimension,
                        "add",
                        True,
                        param_dicts[embedding],
                    ),
                    LinearRegression(),
                    MLkNN(k=2),
                )

        yield EmbeddingClassifier(
            SKLearnEmbedder(SpectralEmbedding(n_components=2)),
            LinearRegression(),
            MLkNN(k=2),
        )

        EmbeddingClassifier(
            CLEMS(metrics.accuracy_score, True), LinearRegression(), MLkNN(k=2), True
        )

    def test_if_embedding_classification_works_on_sparse_input(self):
        for classifier in self.classifiers():
            self.assertClassifierWorksWithSparsity(classifier, "sparse")
            self.assertClassifierPredictsProbabilities(classifier, "sparse")

    def test_if_embedding_classification_works_on_dense_input(self):
        for classifier in self.classifiers():
            self.assertClassifierWorksWithSparsity(classifier, "dense")
            self.assertClassifierPredictsProbabilities(classifier, "dense")

    def test_if_embedding_works_with_cross_validation(self):
        for classifier in self.classifiers():
            self.assertClassifierWorksWithCV(classifier)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from builtins import range
from builtins import zip
from scipy import sparse

from .partition import LabelSpacePartitioningClassifier


class MajorityVotingClassifier(LabelSpacePartitioningClassifier):
    """Majority Voting ensemble classifier

    Divides the label space using provided clusterer class, trains a provided base classifier
    type classifier for each subset and assign a label to an instance
    if more than half of all classifiers (majority) from clusters that contain the label
    assigned the label to the instance.

    Parameters
    ----------
    classifier : :class:`~sklearn.base.BaseEstimator`
        the base classifier that will be used in a class, will be
        automatically put under :code:`self.classifier`.
    clusterer : :class:`~skmultilearn.cluster.LabelSpaceClustererBase`
        object that partitions the output space, will be
        automatically put under :code:`self.clusterer`.
    require_dense : [bool, bool]
        whether the base classifier requires [input, output] matrices
        in dense representation, will be automatically
        put under :code:`self.require_dense`.


    Attributes
    ----------
    model_count_ : int
        number of trained models, in this classifier equal to the number of partitions
    partition_ : List[List[int]], shape=(`model_count_`,)
        list of lists of label indexes, used to index the output space matrix, set in :meth:`_generate_partition`
        via :meth:`fit`
    classifiers : List[:class:`~sklearn.base.BaseEstimator`], shape=(`model_count_`,)
        list of classifiers trained per partition, set in :meth:`fit`


    Examples
    --------
    Here's an example of building an overlapping ensemble of chains

    .. code :: python

        from skmultilearn.ensemble import MajorityVotingClassifier
        from skmultilearn.cluster import FixedLabelSpaceClusterer
        from skmultilearn.problem_transform import ClassifierChain
        from sklearn.naive_bayes import GaussianNB


        classifier = MajorityVotingClassifier(
            clusterer = FixedLabelSpaceClusterer(clusters = [[1,2,3], [0, 2, 5], [4, 5]]),
            classifier = ClassifierChain(classifier=GaussianNB())
        )
        classifier.fit(X_train,y_train)
        predictions = classifier.predict(X_test)

    More advanced examples can be found in `the label relations exploration guide <../labelrelations.ipynb>`_

    """

    def __init__(self, classifier=None, clusterer=None, require_dense=None):
        super(MajorityVotingClassifier, self).__init__(
            classifier=classifier, clusterer=clusterer, require_dense=require_dense
        )

    def predict(self, X):
        """Predict label assignments for X

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse of float
            binary indicator matrix with label assignments with shape
            :code:`(n_samples, n_labels)`
        """
        predictions = [
            self._ensure_input_format(
                self._ensure_input_format(c.predict(X)),
                sparse_format="csc",
                enforce_sparse=True,
            )
            for c in self.classifiers_
        ]

        voters = np.zeros(self._label_count, dtype="int")
        votes = sparse.lil_matrix(
            (predictions[0].shape[0], self._label_count), dtype="int"
        )
        for model in range(self.model_count_):
            for label in range(len(self.partition_[model])):
                votes[:, self.partition_[model][label]] = (
                    votes[:, self.partition_[model][label]]
                    + predictions[model][:, label]
                )
                voters[self.partition_[model][label]] += 1

        nonzeros = votes.nonzero()
        for row, column in zip(nonzeros[0], nonzeros[1]):
            votes[row, column] = np.round(votes[row, column] / float(voters[column]))

        return self._ensure_output_format(votes, enforce_sparse=False)

    def predict_proba(self, X):
        raise NotImplemented(
            "The voting scheme does not define a method for calculating probabilities"
        )



# -*- coding: utf-8 -*-
import numpy as np
import itertools as it


def example_distribution(folds, desired_size):
    """Examples Distribution (ED) measure

     Examples Distribution is a measure of how much a given fold's size deviates from the desired number
     of samples in each of the folds.

    Parameters:
    -----------
    folds : List[List[int]], shape = (n_folds)
        list of indexes of samples assigned per fold

    desired_size : List[int], shape = (n_folds)
        desired number of samples in each fold

    Returns
    -------
    example_distribution_score : float

        The example distribution score
    """
    n_splits = float(len(folds))

    return (
        np.sum(np.fromiter(
            (
                np.abs(len(fold) - desired_fold_size)
                for fold, desired_fold_size in zip(folds, desired_size)
            ),
            dtype=float
        )
        / n_splits)
    )


def get_indicator_representation(row):
    """Convert binary indicator to list of assigned labels

    Parameters:
    -----------

    row : List[{0,1}]
        binary indicator list whether i-th label is assigned or not

    Returns
    -------
    np.array[int]
        list of assigned labels
    """
    return np.where(row != 0)[0]


def get_combination_wise_output_matrix(y, order):
    """Returns label combinations of a given order that are assigned to each row

    Parameters:
    -----------
    y : output matrix or array of arrays (n_samples, n_labels)
        the binary-indicator label assignment per sample representation of the output space

    order : int, >= 1
        the order of label relationship to take into account when balancing sample distribution across labels

    Returns
    -------
    combinations_per_row : List[Set[Tuple[int]]]
        list of combination assignments per row
    """
    return np.array(
        [
            set(
                tuple(combination)
                for combination in it.combinations_with_replacement(
                    get_indicator_representation(row), order
                )
            )
            for row in y
        ]
    )


def get_unique_combinations(combinations_per_row):
    """Performs set.union on a list of sets

    Parameters
    ----------

    combinations_per_row : List[Set[Tuple[int]]]
        list of combination assignments per row

    Returns
    -------
    Set[Tuple[int]]
        all unique label combinations
    """
    if isinstance(combinations_per_row, np.ndarray) and combinations_per_row.ndim > 1:
        combinations_per_row = combinations_per_row.ravel()
    return set.union(*combinations_per_row)


def folds_without_evidence_for_at_least_one_label_combination(y, folds, order=1):
    """Counts the number of folds without evidence for a given Label, Label Pair or Label Combination (FZ, FZLP, FZLC) measure

    A general implementation of FZ - the number of folds that contain at least one label combination of order
    :code:`order` with no positive examples. With :code:`order` = 1, it becomes the FZ measure from Katakis et.al's
    original paper.

    Parameters:
    -----------
    y : output matrix or array of arrays (n_samples, n_labels)
        the binary-indicator label assignment per sample representation of the output space

    folds : List[List[int]], shape = (n_folds)
        list of indexes of samples assigned per fold

    order : int, >= 1
        the order of label relationship to take into account when balancing sample distribution across labels

    Returns
    -------
    score : float
        the number of folds with missing evidence for at least one label combination
    """
    combinations_per_row = get_combination_wise_output_matrix(y, order)
    all_combinations = get_unique_combinations(combinations_per_row)
    return np.sum(
        [
            get_unique_combinations(combinations_per_row[[fold]]) != all_combinations
            for fold in folds
        ]
    )


def folds_label_combination_pairs_without_evidence(y, folds, order):
    """Fold - Label / Label Pair / Label Combination (FLZ, FLPZ, FLCZ)  pair count measure

    A general implementation of FLZ - the number of pairs of fold and label combination of a given order for which
    there is no positive evidence in that fold for that combination. With :code:`order` = 1, it becomes the FLZ
    measure from Katakis et.al's original paper, with :code:`order` = 2, it becomes the FLPZ measure from
    Szymański et. al.'s paper.

    Parameters:
    -----------
    y : output matrix or array of arrays (n_samples, n_labels)
        the binary-indicator label assignment per sample representation of the output space

    folds : List[List[int]], shape = (n_folds)
        list of indexes of samples assigned per fold

    order : int, >= 1
        the order of label relationship to take into account when balancing sample distribution across labels

    Returns
    -------
    score : float
        the number of fold-label combination pairs with missing evidence
    """
    combinations_per_row = get_combination_wise_output_matrix(y, order)
    all_combinations = get_unique_combinations(combinations_per_row)
    return np.sum(
        [
            len(
                all_combinations.difference(
                    get_unique_combinations(combinations_per_row[[fold]])
                )
            )
            for fold in folds
        ]
    )


def percentage_of_label_combinations_without_evidence_per_fold(y, folds, order):
    """Fold - Label / Label Pair / Label Combination (FLZ, FLPZ, FLCZ)  pair count measure

    A general implementation of FLZ - the number of pairs of fold and label combination of a given order for which
    there is no positive evidence in that fold for that combination. With :code:`order` = 1, it becomes the FLZ
    measure from Katakis et.al's original paper, with :code:`order` = 2, it becomes the FLPZ measure from
    Szymański et. al.'s paper.

    Parameters:
    -----------
    y : output matrix or array of arrays (n_samples, n_labels)
        the binary-indicator label assignment per sample representation of the output space

    folds : List[List[int]], shape = (n_folds)
        list of indexes of samples assigned per fold

    order : int, >= 1
        the order of label relationship to take into account when balancing sample distribution across labels

    Returns
    -------
    score : float
        the number of fold-label combination pairs with missing evidence
    """
    combinations_per_row = get_combination_wise_output_matrix(y, order)
    all_combinations = get_unique_combinations(combinations_per_row)
    number_of_combinations = float(len(all_combinations))
    return [
        1.0
        - len(get_unique_combinations(combinations_per_row[[fold]]))
        / number_of_combinations
        for fold in folds
    ]


def label_combination_distribution(y, folds, order):
    """Label / Label Pair / Label Combination Distribution (LD, LPD, LCZD) measure

    A general implementation of Label / Label Pair / Label Combination Distribution - a measure that evaluates
    how the proportion of positive evidence for a label / label pair / label combination to the negative evidence
    for a label (pair/combination) deviates from the same proportion in the entire data set, averaged over all folds and labels.

    With :code:`order` = 1, it becomes the LD measure from Katakis et.al's original paper, with :code:`order` = 2, it
    becomes the LPD measure from Szymański et. al.'s paper.

    Parameters:
    -----------
    y : output matrix or array of arrays (n_samples, n_labels)
        the binary-indicator label assignment per sample representation of the output space

    folds : List[List[int]], shape = (n_folds)
        list of indexes of samples assigned per fold

    order : int, >= 1
        the order of label relationship to take into account when balancing sample distribution across labels

    Returns
    -------
    score : float
        the label / label pair / label combination distribution score
    """

    def _get_proportion(x, y):
        return y / float(x - y)

    combinations_per_row = get_combination_wise_output_matrix(y, order)
    all_combinations = get_unique_combinations(combinations_per_row)
    number_of_samples = y.shape[0]
    number_of_combinations = float(len(all_combinations))
    number_of_folds = float(len(folds))

    external_sum = 0
    for combination in all_combinations:
        number_of_samples_with_combination = np.sum(
            [
                1
                for combinations_in_row in combinations_per_row
                if combination in combinations_in_row
            ]
        )

        d = _get_proportion(number_of_samples, number_of_samples_with_combination)
        internal_sum = 0
        for fold in folds:
            S_i_j = np.sum(
                [
                    1
                    for combinations_in_row in combinations_per_row[fold]
                    if combination in combinations_in_row
                ]
            )
            fold_size = len(fold)
            s = _get_proportion(fold_size, S_i_j)
            internal_sum += np.abs(s - d)

        internal_sum /= number_of_folds
        external_sum += internal_sum

    return external_sum / number_of_combinations



"""
The :mod:`skmultilearn.cluster` module gathers label space clustering methods.


+-------------------------------------------------------------+----------------------------------------------------------+
| Name                                                        | Description                                              |
+=============================================================+==========================================================+
| :class:`~skmultilearn.cluster.FixedLabelSpaceClusterer`     | Return a predefined fixed clustering, usually driven by  |
|                                                             | expert knowledge                                         |
+-------------------------------------------------------------+----------------------------------------------------------+
| :class:`~skmultilearn.cluster.MatrixLabelSpaceClusterer`    | Cluster the label space using a scikit-compatible        |
|                                                             | matrix-based clusterer                                   |
+-------------------------------------------------------------+----------------------------------------------------------+
| :class:`~skmultilearn.cluster.GraphToolLabelGraphClusterer` | Fits a Stochastic Block Model to the Label Graph and     |
|                                                             | infers the communities                                   |
+-------------------------------------------------------------+----------------------------------------------------------+
| :class:`~skmultilearn.cluster.StochasticBlockModel`         | A Stochastic Blockmodel class                            |
+-------------------------------------------------------------+----------------------------------------------------------+
| :class:`~skmultilearn.cluster.IGraphLabelGraphClusterer`    | Clusters label space using igraph community detection    |
+-------------------------------------------------------------+----------------------------------------------------------+
| :class:`~skmultilearn.cluster.RandomLabelSpaceClusterer`    | Randomly divides label space into equally-sized clusters |
+-------------------------------------------------------------+----------------------------------------------------------+
| :class:`~skmultilearn.cluster.NetworkXLabelGraphClusterer`  | Cluster label space with NetworkX community detection    |
+-------------------------------------------------------------+----------------------------------------------------------+

"""

from .base import LabelCooccurrenceGraphBuilder
from .fixed import FixedLabelSpaceClusterer
from .matrix import MatrixLabelSpaceClusterer
from .random import RandomLabelSpaceClusterer

__all__ = [
    "FixedLabelSpaceClusterer",
    "LabelCooccurrenceGraphBuilder",
    "MatrixLabelSpaceClusterer",
    "RandomLabelSpaceClusterer",
]

# graphtool import optional (is GPL-ed, does not work on windows)
try:
    from .graphtool import GraphToolLabelGraphClusterer, StochasticBlockModel

    __all__ += ["GraphToolLabelGraphClusterer", "StochasticBlockModel"]
except ImportError:
    pass

# python-igraph import optional (is GPL-ed)
try:
    from .igraph import IGraphLabelGraphClusterer

    __all__ += ["IGraphLabelGraphClusterer"]
except ImportError:
    pass

# networkx import is optional
try:
    from .networkx import NetworkXLabelGraphClusterer

    __all__ += ["NetworkXLabelGraphClusterer"]
except ImportError:
    pass



from .voting import MajorityVotingClassifier
from ..cluster.random import RandomLabelSpaceClusterer
from ..problem_transform import LabelPowerset
from ..base import MLClassifierBase


class RakelO(MLClassifierBase):
    """Overlapping RAndom k-labELsets multi-label classifier

    Divides the label space in to m subsets of size k, trains a Label Powerset
    classifier for each subset and assign a label to an instance
    if more than half of all classifiers (majority) from clusters that contain the label
    assigned the label to the instance.

    Parameters
    ----------
    base_classifier: :class:`~sklearn.base.BaseEstimator`
        scikit-learn compatible base classifier, will be set under `self.classifier.classifier`.
    base_classifier_require_dense : [bool, bool]
        whether the base classifier requires [input, output] matrices
        in dense representation. Will be automatically
        set under `self.classifier.require_dense`
    labelset_size : int
        the desired size of each of the partitions, parameter k according to paper.
        According to paper, the best parameter is 3, so it's set as default
        Will be automatically set under `self.labelset_size`
    model_count : int
        the desired number of classifiers, parameter m according to paper.
        According to paper, the best value for this parameter is 2M (being M the number of labels)
        Will be automatically set under :code:`self.model_count_`.


    Attributes
    ----------
    classifier : :class:`~skmultilearn.ensemble.MajorityVotingClassifier`
        the voting classifier initialized with :class:`~skmultilearn.problem_transform.LabelPowerset` multi-label
        classifier with `base_classifier` and :class:`~skmultilearn.cluster.random.RandomLabelSpaceClusterer`


    References
    ----------

    If you use this class please cite the paper introducing the method:

    .. code :: latex

        @ARTICLE{5567103,
            author={G. Tsoumakas and I. Katakis and I. Vlahavas},
            journal={IEEE Transactions on Knowledge and Data Engineering},
            title={Random k-Labelsets for Multilabel Classification},
            year={2011},
            volume={23},
            number={7},
            pages={1079-1089},
            doi={10.1109/TKDE.2010.164},
            ISSN={1041-4347},
            month={July},
        }

    Examples
    --------

    Here's a simple example of how to use this class with a base classifier from scikit-learn to teach 6 classifiers
    each trained on a quarter of labels, which is sure to overlap:

    .. code :: python

        from sklearn.naive_bayes import GaussianNB
        from skmultilearn.ensemble import RakelO

        classifier = RakelO(
            base_classifier=GaussianNB(),
            base_classifier_require_dense=[True, True],
            labelset_size=y_train.shape[1] // 4,
            model_count_=6
        )

        classifier.fit(X_train, y_train)
        prediction = classifier.predict(X_train, y_train)

    """

    def __init__(
        self,
        base_classifier=None,
        model_count=None,
        labelset_size=3,
        base_classifier_require_dense=None,
    ):
        super(RakelO, self).__init__()

        self.model_count = model_count
        self.labelset_size = labelset_size
        self.base_classifier = base_classifier
        self.base_classifier_require_dense = base_classifier_require_dense
        self.copyable_attrs = [
            "model_count",
            "labelset_size",
            "base_classifier_require_dense",
            "base_classifier",
        ]

    def fit(self, X, y):
        """Fits classifier to training data

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self
        """
        self.classifier = MajorityVotingClassifier(
            classifier=LabelPowerset(
                classifier=self.base_classifier,
                require_dense=self.base_classifier_require_dense,
            ),
            clusterer=RandomLabelSpaceClusterer(
                cluster_size=self.labelset_size,
                cluster_count=self.model_count,
                allow_overlap=True,
            ),
            require_dense=[False, False],
        )
        return self.classifier.fit(X, y)

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """

        return self.classifier.predict(X)

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `float in [0.0, 1.0]`, shape=(n_samples, n_labels)
            matrix with label assignment probabilities
        """
        return self.classifier.predict_proba(X)



from __future__ import absolute_import

import numpy as np
import random

from .base import LabelSpaceClustererBase
from .helpers import _euclidean_distance, _recalculateCenters, _countNumberOfAparitions


class BalancedKMeansClusterer(LabelSpaceClustererBase):
    """Cluster the label space regarding the algorithm of balancedkMeans, used by HOMER"""

    def __init__(self, k=None, it=None):
        """Initializes the clusterer

        Attributes
        ----------
        k: int
                Number of partitions to be made to the label-space
        it: int
                Number of iterations for the algorithm to find the best neighbours
        """
        super(BalancedKMeansClusterer, self).__init__()
        self.k = k
        self.it = it

    def fit_predict(self, X, y):
        """Performs clustering on y and returns list of label lists

        Builds a label list taking care of the distance between labels

        Parameters
        ----------
        X : currently unused, left for scikit compatibility
        y : scipy.sparse
            label space of shape :code:`(n_samples, n_labels)`

        Returns
        -------
        array of arrays
            numpy array of arrays of label indexes, where each sub-array
            represents labels that are in a separate community
        """
        number_of_labels = y.shape[1]
        # Assign a label to a cluster no. label ordinal %  number of labeSls
        # We have to do the balance k-means and then use it for HOMER with the label powerset
        Centers = []
        y = y.todense()
        for i in range(0, self.k):
            auxVector = y[:, random.randint(0, number_of_labels - 1)]
            Centers.append(np.asarray(auxVector))
        # Now we have the clusters created and we need to make each label its corresponding cluster

        while self.it > 0:
            balancedCluster = []
            for j in range(0, number_of_labels):
                auxVector = y[:, j]
                v = np.asarray(auxVector)
                # Now we calculate the distance and store it in an array
                distances = []
                for i in range(0, self.k):
                    # Store the distances
                    distances.append(_euclidean_distance(v, Centers[i]))
                finished = False
                while not finished:
                    minIndex = np.argmin(distances)
                    balancedCluster.append(minIndex)
                    # Now we have the cluster we want to add this label to
                    numberOfAparitions = _countNumberOfAparitions(
                        balancedCluster, minIndex
                    )
                    if float(numberOfAparitions) > (
                        float(float(number_of_labels) / float(self.k)) + 1
                    ):
                        distances[minIndex] = float("inf")
                        balancedCluster.pop()
                    else:
                        finished = True

            Centers = _recalculateCenters(np.asarray(y), balancedCluster, self.k)
            self.it = self.it - 1

        # Returns a list of list with the clusterers
        labelCluster = []
        for i in range(0, self.k):
            cluster = []
            for j in range(0, len(balancedCluster)):
                if int(i) == int(balancedCluster[j]):
                    cluster.append(int(j))
            labelCluster.append(cluster)

        return np.asarray(labelCluster, dtype=object)



import math
import numpy as np


def _membership_to_list_of_communities(membership_vector, size):
    """Convert membership vector to list of lists of vertices in each community

    Parameters
    ----------
    membership_vector : list of int
        community membership i.e. vertex/label `i` is in community `membership_vector[i]`
    size : int
        the number of communities present in the membership vector


    Returns
    -------
    list_of_members : list of lists of int
        list of lists of vertex/label ids in each community per community
    """
    list_of_members = [[] for _ in range(size)]
    for vertex_id, community_id in enumerate(membership_vector):
        list_of_members[community_id].append(vertex_id)
    return list_of_members


def _overlapping_membership_to_list_of_communities(membership_vector, size):
    """Convert membership vector to list of lists of vertices/labels in each community

    Parameters
    ----------
    membership_vector : list of lists of int
        community membership i.e. vertex/label `i` is in communities
        from list `membership_vector[i]`
    size : int
        the number of communities present in the membership vector


    Returns
    -------
    list_of_members : list of lists of int
        list of lists of vertex/label ids in each community per community
    """
    list_of_members = [[] for _ in range(size)]
    for vertex_id, community_ids in enumerate(membership_vector):
        for community_id in community_ids:
            list_of_members[community_id].append(vertex_id)
    return list_of_members


def _euclidean_distance(array1, array2):
    """Returns the euclidean distance of two arrays

    Parameters
    ----------
    array1 : array of numbers

    array2 : array of numbers

    Returns
    -------
    distance : float
        float with the euclidean distance, False if not possible
    """
    # Ensure that both arrays hava the same length
    if len(array1) != len(array2):
        return False
    else:
        distance = 0.0
        for i in range(0, len(array1)):
            distance = distance + pow(array1[i] - array2[i], 2)
        distance = math.sqrt(distance)
        return distance


def _recalculateCenters(y, balancedCluster, k):
    Centers = []
    kAux = 0
    while kAux < k:
        vectorAux = np.zeros(len(y))
        for i in range(0, len(balancedCluster)):
            if int(kAux) == int(balancedCluster[i]):
                # We have to fill the vector
                for j in range(0, len(y)):
                    vectorAux[j] += y[j, i]

        vectorAux /= k
        Centers.append(vectorAux)
        kAux += 1
    return Centers


def _countNumberOfAparitions(array, number):
    """Number of aparitions of a number in an array

    Parameters
    ----------
    array : array of numbers

    number : number to search for

    Returns
    -------
    aparaitions : int
        Number of aparitions of the number in the given array
    """
    aparitions = 0
    for i in range(0, len(array)):
        if array[i] == number:
            aparitions += 1
    return aparitions



# -*- coding: utf-8 -*-
"""IGraph based Label Space Clustering"""
from __future__ import absolute_import

import numpy as np
from builtins import range
from igraph import Graph

from .base import LabelGraphClustererBase


class IGraphLabelGraphClusterer(LabelGraphClustererBase):
    """Clusters the label space using igraph community detection methods

    This clusterer constructs an igraph representation of the Label Graph generated by graph builder and detects
    communities in it using community detection methods from the igraph library. Detected communities are converted to
    a label space clustering. The approach has been described in `this paper concerning data-driven label space division
    <http://www.mdpi.com/1099-4300/18/8/282/htm>`_.

    Parameters
    ----------
    graph_builder: a GraphBuilderBase inherited transformer
        the graph builder to provide the adjacency matrix and weight map for the underlying graph
    method: string
        the community detection method to use, this clusterer supports the following community detection methods:

        +----------------------+---------------------------------------------------------------------------------------+
        | Method name string   |                             Description                                               |
        +======================+=======================================================================================+
        | fastgreedy_          | Detecting communities with largest modularity using incremental greedy search         |
        +----------------------+---------------------------------------------------------------------------------------+
        | infomap_             | Detecting communities through information flow compressing simulated via random walks |
        +----------------------+---------------------------------------------------------------------------------------+
        | label_propagation_   | Detecting communities from colorings via multiple label propagation on the graph      |
        +----------------------+---------------------------------------------------------------------------------------+
        | leading_eigenvector_ | Detecting communities with largest modularity through adjacency matrix eigenvectors   |
        +----------------------+---------------------------------------------------------------------------------------+
        | multilevel_          | Recursive communitiy detection with largest modularity step by step maximization      |
        +----------------------+---------------------------------------------------------------------------------------+
        | walktrap_            |  Finding communities by trapping many random walks                                    |
        +----------------------+---------------------------------------------------------------------------------------+

        .. _fastgreedy: http://igraph.org/python/doc/igraph.Graph-class.html#community_fastgreedy
        .. _infomap: http://igraph.org/python/doc/igraph.Graph-class.html#community_infomap
        .. _label_propagation: http://igraph.org/python/doc/igraph.Graph-class.html#community_label_propagation
        .. _leading_eigenvector: http://igraph.org/python/doc/igraph.Graph-class.html#community_leading_eigenvector
        .. _multilevel: http://igraph.org/python/doc/igraph.Graph-class.html#community_multilevel
        .. _walktrap: http://igraph.org/python/doc/igraph.Graph-class.html#community_walktrap



    Attributes
    ----------
    graph_ : igraph.Graph
        the igraph Graph object containing the graph representation of graph builder's adjacency matrix and weights
    weights_ : { 'weight' : list of values in edge order of graph edges }
        edge weights stored in a format recognizable by the igraph module


    .. note ::

        This clusterer is GPL-licenced and will taint your code with GPL restrictions.


    References
    ----------

    If you use this clusterer please cite the igraph paper and the clustering paper:

    .. code:: latex

        @Article{igraph,
            title = {The igraph software package for complex network research},
            author = {Gabor Csardi and Tamas Nepusz},
            journal = {InterJournal},
            volume = {Complex Systems},
            pages = {1695},
            year = {2006},
            url = {http://igraph.org},
        }

        @Article{datadriven,
            author = {Szymański, Piotr and Kajdanowicz, Tomasz and Kersting, Kristian},
            title = {How Is a Data-Driven Approach Better than Random Choice in
            Label Space Division for Multi-Label Classification?},
            journal = {Entropy},
            volume = {18},
            year = {2016},
            number = {8},
            article_number = {282},
            url = {http://www.mdpi.com/1099-4300/18/8/282},
            issn = {1099-4300},
            doi = {10.3390/e18080282}
        }


    Examples
    --------

    An example code for using this clusterer with a classifier looks like this:

    .. code-block:: python

        from sklearn.ensemble import RandomForestClassifier
        from skmultilearn.problem_transform import LabelPowerset
        from skmultilearn.cluster import IGraphLabelGraphClusterer, LabelCooccurrenceGraphBuilder
        from skmultilearn.ensemble import LabelSpacePartitioningClassifier

        # construct base forest classifier
        base_classifier = RandomForestClassifier(n_estimators=1000)

        # construct a graph builder that will include
        # label relations weighted by how many times they
        # co-occurred in the data, without self-edges
        graph_builder = LabelCooccurrenceGraphBuilder(
            weighted = True,
            include_self_edges = False
        )

        # setup problem transformation approach with sparse matrices for random forest
        problem_transform_classifier = LabelPowerset(classifier=base_classifier,
            require_dense=[False, False])

        # setup the clusterer to use, we selected the fast greedy modularity-maximization approach
        clusterer = IGraphLabelGraphClusterer(graph_builder=graph_builder, method='fastgreedy')

        # setup the ensemble metaclassifier
        classifier = LabelSpacePartitioningClassifier(problem_transform_classifier, clusterer)

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)

    For more use cases see `the label relations exploration guide <../labelrelations.ipynb>`_.

    """

    _METHODS = {
        "fastgreedy": lambda graph, w=None: graph.community_fastgreedy(
            weights=w
        ).as_clustering(),
        "infomap": lambda graph, w=None: graph.community_infomap(edge_weights=w),
        "label_propagation": lambda graph, w=None: graph.community_label_propagation(
            weights=w
        ),
        "leading_eigenvector": lambda graph, w=None: graph.community_leading_eigenvector(
            weights=w
        ),
        "multilevel": lambda graph, w=None: graph.community_multilevel(weights=w),
        "walktrap": lambda graph, w=None: graph.community_walktrap(
            weights=w
        ).as_clustering(),
    }

    def __init__(self, graph_builder, method):
        super(IGraphLabelGraphClusterer, self).__init__(graph_builder)
        self.method = method

        if method not in IGraphLabelGraphClusterer._METHODS:
            raise ValueError(
                "{} not a supported igraph community detection method".format(method)
            )

    def fit_predict(self, X, y):
        """Performs clustering on y and returns list of label lists

        Builds a label graph using the provided graph builder's `transform` method
        on `y` and then detects communities using the selected `method`.

        Sets :code:`self.weights_` and :code:`self.graph_`.

        Parameters
        ----------
        X : None
            currently unused, left for scikit compatibility
        y : scipy.sparse
            label space of shape :code:`(n_samples, n_labels)`

        Returns
        -------
        arrray of arrays of label indexes (numpy.ndarray)
            label space division, each sublist represents labels that are in that community
        """
        edge_map = self.graph_builder.transform(y)

        if self.graph_builder.is_weighted:
            self.weights_ = dict(weight=list(edge_map.values()))
        else:
            self.weights_ = dict(weight=None)

        self.graph_ = Graph(
            edges=[x for x in edge_map],
            vertex_attrs=dict(name=list(range(1, y.shape[1] + 1))),
            edge_attrs=self.weights_,
        )

        return np.array(
            IGraphLabelGraphClusterer._METHODS[self.method](
                self.graph_, self.weights_["weight"]
            ),
            dtype=object
        )



from __future__ import absolute_import
from __future__ import print_function

import graph_tool.all as gt
import numpy as np

from .base import LabelGraphClustererBase
from .helpers import (
    _membership_to_list_of_communities,
    _overlapping_membership_to_list_of_communities,
)


class StochasticBlockModel:
    """A Stochastic Blockmodel fit to Label Graph

    This contains a stochastic block model instance constructed for a block model variant specified in parameters.
    It can be fit to an instance of a graph and set of weights. More information on how to select parameters can be
    found in `the extensive introduction into Stochastic Block Models
    <https://graph-tool.skewed.de/static/doc/demos/inference/inference.html>`_ in graphtool documentation.

    Parameters
    ----------
    nested: boolean
        whether to build a nested Stochastic Block Model or the regular variant,
        will be automatically put under :code:`self.nested`.
    use_degree_correlation: boolean
        whether to correct for degree correlation in modeling, will be automatically
        put under :code:`self.use_degree_correlation`.
    allow_overlap: boolean
        whether to allow overlapping clusters or not, will be automatically
        put under :code:`self.allow_overlap`.
    weight_model: string or None
        decide whether to generate a weighted or unweighted graph,
        will be automatically put under :code:`self.weight_model`.

    Attributes
    ----------
    model_: graph_tool.inference.BlockState or its subclass
        an instance of the fitted model obtained from graph-tool
    """

    def __init__(self, nested, use_degree_correlation, allow_overlap, weight_model):
        self.nested = nested
        self.use_degree_correlation = use_degree_correlation
        self.allow_overlap = allow_overlap
        self.weight_model = weight_model
        self.model_ = None

    def fit_predict(self, graph, weights):
        """Fits model to a given graph and weights list

        Sets :code:`self.model_` to the state of graphtool's Stochastic Block Model the after fitting.

        Attributes
        ----------
        graph: graphtool.Graph
            the graph to fit the model to
        weights: graphtool.EdgePropertyMap<double>
            the property map: edge -> weight (double) to fit the model to, if weighted variant
            is selected

        Returns
        -------
        numpy.ndarray
            partition of labels, each sublist contains label indices
            related to label positions in :code:`y`
        """
        if self.weight_model:
            self.model_ = self._model_fit_function()(
                graph,
                deg_corr=self.use_degree_correlation,
                overlap=self.allow_overlap,
                state_args=dict(recs=[weights], rec_types=[self.weight_model]),
            )
        else:
            self.model_ = self._model_fit_function()(
                graph, deg_corr=self.use_degree_correlation, overlap=self.allow_overlap
            )
        return self._detect_communities()

    def _detect_communities(self):
        if self.nested:
            lowest_level = self.model_.get_levels()[0]
        else:
            lowest_level = self.model_

        number_of_communities = lowest_level.get_B()
        if self.allow_overlap:
            # the overlaps block returns
            # membership vector, and also edges vectors, we need just the membership here at the moment
            membership_vector = list(lowest_level.get_overlap_blocks()[0])
        else:
            membership_vector = list(lowest_level.get_blocks())

        if self.allow_overlap:
            return _overlapping_membership_to_list_of_communities(
                membership_vector, number_of_communities
            )

        return _membership_to_list_of_communities(
            membership_vector, number_of_communities
        )

    def _model_fit_function(self):
        if self.nested:
            return gt.minimize_nested_blockmodel_dl
        else:
            return gt.minimize_blockmodel_dl


class GraphToolLabelGraphClusterer(LabelGraphClustererBase):
    """Fits a Stochastic Block Model to the Label Graph and infers the communities

    This clusterer clusters the label space using by fitting a stochastic block
    model to the label network and inferring the community structure using graph-tool.
    The obtained community structure is returned as the label clustering. More information on the inference itself
    can be found in `the extensive introduction into Stochastic Block Models
    <https://graph-tool.skewed.de/static/doc/demos/inference/inference.html>`_ in graphtool documentation.

    Parameters
    ----------
    graph_builder: a GraphBuilderBase inherited transformer
        the graph builder to provide the adjacency matrix and weight map for the underlying graph
    model: StochasticBlockModel
        the desired stochastic block model variant to use


    Attributes
    ----------
    graph_ : graphtool.Graph
        object representing a label co-occurence graph
    weights_ : graphtool.EdgeProperty<double>
        edge weights defined by graph builder stored in a graphtool compatible format


    .. note ::

        This functionality is still undergoing research.


    .. note ::

        This clusterer is GPL-licenced and will taint your code with GPL restrictions.



    References
    ----------

    If you use this class please cite:

    .. code : latex

        article{peixoto_graph-tool_2014,
         title = {The graph-tool python library},
         url = {http://figshare.com/articles/graph_tool/1164194},
         doi = {10.6084/m9.figshare.1164194},
         urldate = {2014-09-10},
         journal = {figshare},
         author = {Peixoto, Tiago P.},
         year = {2014},
         keywords = {all, complex networks, graph, network, other}}


    Examples
    --------

    An example code for using this clusterer with a classifier looks like this:

    .. code-block:: python

        from sklearn.ensemble import RandomForestClassifier
        from skmultilearn.problem_transform import LabelPowerset
        from skmultilearn.cluster import IGraphLabelGraphClusterer, LabelCooccurrenceGraphBuilder
        from skmultilearn.ensemble import LabelSpacePartitioningClassifier

        # construct base forest classifier
        base_classifier = RandomForestClassifier(n_estimators=1000)

        # construct a graph builder that will include
        # label relations weighted by how many times they
        # co-occurred in the data, without self-edges
        graph_builder = LabelCooccurrenceGraphBuilder(
            weighted = True,
            include_self_edges = False
        )

        # select parameters for the model, we fit a flat,
        # non-degree correlated, partitioning model
        # which will use fit the normal distribution as the weights model
        model = StochasticBlockModel(
            nested=False,
            use_degree_correlation=True,
            allow_overlap=False,
            weight_model='real-normal'
        )

        # setup problem transformation approach with sparse matrices for random forest
        problem_transform_classifier = LabelPowerset(classifier=base_classifier,
            require_dense=[False, False])

        # setup the clusterer to use, we selected the fast greedy modularity-maximization approach
        clusterer = GraphToolLabelGraphClusterer(graph_builder=graph_builder, model=model)

        # setup the ensemble metaclassifier
        classifier = LabelSpacePartitioningClassifier(problem_transform_classifier, clusterer)

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)

    For more use cases see `the label relations exploration guide <../labelrelations.ipynb>`_.

    """

    def __init__(self, graph_builder, model):
        super(GraphToolLabelGraphClusterer, self).__init__(graph_builder)

        self.model = model
        self.graph_builder = graph_builder

    def fit_predict(self, X, y):
        """Performs clustering on y and returns list of label lists

        Builds a label graph using the provided graph builder's `transform` method
        on `y` and then detects communities using the selected `method`.

        Sets :code:`self.weights_` and :code:`self.graph_`.

        Parameters
        ----------
        X : None
            currently unused, left for scikit compatibility
        y : scipy.sparse
            label space of shape :code:`(n_samples, n_labels)`

        Returns
        -------
        arrray of arrays of label indexes (numpy.ndarray)
            label space division, each sublist represents labels that are in that community
        """
        self._build_graph_instance(y)
        clusters = self.model.fit_predict(self.graph_, weights=self.weights_)
        return np.array([community for community in clusters if len(community) > 0])

    def _build_graph_instance(self, y):
        edge_map = self.graph_builder.transform(y)

        g = gt.Graph(directed=False)
        g.add_vertex(y.shape[1])

        self.weights_ = g.new_edge_property("double")

        for edge, weight in edge_map.items():
            e = g.add_edge(edge[0], edge[1])
            self.weights_[e] = weight

        self.graph_ = g
